from typing import Tuple
import torch


# Broadcast Utilities
def normalize_bcast_dims(*shapes):
    """
    Normalize the lengths of the input shapes to have the same length.
    The shapes are padded at the front by 1 to make the lengths equal.

    Examples
    --------
    >>> normalize_bcast_dims([1, 2, 3], [2, 3])
    [[1, 2, 3], [1, 2, 3]]

    Parameters
    ----------
    shapes: List[List[int]]
        The shapes to normalize.

    Returns
    -------
    List[List[int]]
        The normalized shapes.

    """
    maxlens = max([len(shape) for shape in shapes])
    res = [[1] * (maxlens - len(shape)) + list(shape) for shape in shapes]
    return res


def get_bcasted_dims(*shapes):
    """Return the broadcasted shape of the given shapes.

    Examples
    --------
    >>> get_bcasted_dims([1, 2, 5], [2, 3, 4])
    [2, 3, 5]

    Parameters
    ----------
    shapes: List[List[int]]
        The shapes to broadcast.

    Returns
    -------
    List[int]
        The broadcasted shape.

    """
    shapes = normalize_bcast_dims(*shapes)
    return [max(*a) for a in zip(*shapes)]


def match_dim(*xs: torch.Tensor,
              contiguous: bool = False) -> Tuple[torch.Tensor, ...]:
    """match the N-1 dimensions of x and xq for searchsorted and gather with dim=-1

    Examples
    --------
    >>> x = torch.randn(10, 5)
    >>> xq = torch.randn(10, 3)
    >>> x_new, xq_new = match_dim(x, xq)
    >>> x_new.shape
    torch.Size([10, 5])
    >>> xq_new.shape
    torch.Size([10, 3])

    """
    orig_shapes = tuple(x.shape[:-1] for x in xs)
    shape = tuple(get_bcasted_dims(*orig_shapes))
    xs_new = tuple(x.expand(shape + (x.shape[-1],)) for x in xs)
    if contiguous:
        xs_new = tuple(x.contiguous() for x in xs_new)
    return xs_new



import pytest
try:
    import torch
    from deepchem.utils.differentiation_utils import EditableModule
    has_torch = True
except ModuleNotFoundError:
    has_torch = False


@pytest.mark.torch
def test_getparams():

    class A(EditableModule):

        def __init__(self, a):
            self.b = a * a

        def mult(self, x):
            return self.b * x

        def getparamnames(self, methodname, prefix=""):
            if methodname == "mult":
                return [prefix + "b"]
            else:
                raise KeyError()

    a = torch.tensor(2.0)
    x = torch.tensor(0.4)
    alpha = A(a)
    assert alpha.mult(x) == torch.tensor(1.6)
    assert alpha.getparams("mult") == [torch.tensor(4.)]


@pytest.mark.torch
def test_setparams():

    class A(EditableModule):

        def __init__(self, a):
            self.b = a * a

        def mult(self, x):
            return self.b * x

        def getparamnames(self, methodname, prefix=""):
            if methodname == "mult":
                return [prefix + "b"]
            else:
                raise KeyError()

    a = torch.tensor(2.0)
    x = torch.tensor(4.0)
    alpha = A(a)
    assert alpha.mult(x) == torch.tensor(16)
    alpha.setparams("mult", torch.tensor(5.0))
    assert alpha.mult(x) == torch.tensor(20.0)


@pytest.mark.torch
def test_cached_getparamnames():

    class A(EditableModule):

        def __init__(self, a):
            self.b = a * a

        def mult(self, x):
            return self.b * x

        def getparamnames(self, methodname, prefix=""):
            if methodname == "mult":
                return [prefix + "b"]
            else:
                raise KeyError()

    a = torch.tensor(2.0)
    alpha = A(a)
    assert alpha.cached_getparamnames("mult") == ["b"]


@pytest.mark.torch
def test_getuniqueparams():

    class A(EditableModule):

        def __init__(self, a):
            self.b = a * a

        def mult(self, x):
            return self.b**2 * x

        def getparamnames(self, methodname, prefix=""):
            if methodname == "mult":
                return [prefix + "b"]
            else:
                raise KeyError()

    a = torch.tensor(2.0)
    x = torch.tensor(0.4)
    alpha = A(a)
    assert alpha.mult(x) == torch.tensor(6.4)
    assert alpha.getuniqueparams("mult") == [torch.tensor(4.)]  # Not 16.0


@pytest.mark.torch
def test_setuniqueparams():

    class A(EditableModule):

        def __init__(self, a):
            self.b = a * a

        def mult(self, x):
            return self.b**2 * x

        def getparamnames(self, methodname, prefix=""):
            if methodname == "mult":
                return [prefix + "b"]
            else:
                raise KeyError()

    a = torch.tensor(2.0)
    x = torch.tensor(0.4)
    alpha = A(a)
    assert alpha.mult(x) == torch.tensor(6.4)
    assert alpha.getuniqueparams("mult") == [torch.tensor(4.)]
    alpha.setuniqueparams("mult", torch.tensor(5.0))
    assert alpha.mult(x) == torch.tensor(10.0)


@pytest.mark.torch
def test_get_unique_params_idxs():

    class A(EditableModule):

        def __init__(self, a):
            self.b = a * a
            self.c = a * a * a

        def mult(self, x):
            return self.b * self.c * x

        def getparamnames(self, methodname, prefix=""):
            if methodname == "mult":
                return [prefix + "b", prefix + "c"]
            else:
                raise KeyError()

    a = torch.tensor(2.0)
    x = torch.tensor(4.0)
    alpha = A(a)
    assert alpha.mult(x) == torch.tensor(128.0)
    assert alpha.getparams("mult") == [torch.tensor(4.), torch.tensor(8.)]
    assert alpha._get_unique_params_idxs("mult") == [0, 1]


@pytest.mark.torch
def test_assertparams():
    """Test that assertparams works correctly.
    also checks the private methods as they are used in it.
    - __assert_method_preserve
    - __assert_get_correct_params
    - __list_operating_params

    """

    class A(EditableModule):

        def __init__(self, a):
            self.b = a * a
            self.c = a * a * a

        def mult(self, x):
            return self.b * self.c * x

        def getparamnames(self, methodname, prefix=""):
            if methodname == "mult":
                return [prefix + "b", prefix + "c"]
            else:
                raise KeyError()

    a = torch.tensor(2.0)
    x = torch.tensor(4.0)
    alpha = A(a)
    assert alpha.mult(x) == torch.tensor(128.0)
    assert alpha.getparams("mult") == [torch.tensor(4.), torch.tensor(8.)]
    alpha.assertparams(alpha.mult, x)


@pytest.mark.torch
def test_getparamnames():

    class A(EditableModule):

        def __init__(self, a):
            self.b = a * a

        def mult(self, x):
            return self.b * x

        def getparamnames(self, methodname, prefix=""):
            if methodname == "mult":
                return [prefix + "b"]
            else:
                raise KeyError()

    a = torch.tensor(2.0).requires_grad_()
    x = torch.tensor(0.4).requires_grad_()
    alpha = A(a)
    assert alpha.mult(x) == torch.tensor(1.6000)
    assert alpha.getparamnames("mult") == ['b']


@pytest.mark.torch
def test_traverse_obj():
    from deepchem.utils.differentiation_utils.editable_module import _traverse_obj, torch_float_type

    class A:

        def __init__(self):
            self.a = 2
            self.b = torch.tensor(3.0)
            self.c = torch.tensor(4.0)
            self.d = torch.tensor(5.0)

    a = A()

    def action(elmt, name, objdict, key):
        print(name, elmt)

    def crit(elmt):
        return isinstance(elmt, torch.Tensor) and elmt.dtype in torch_float_type

    a = _traverse_obj(a, "", action, crit)  # Check Doesn't Crashes


@pytest.mark.torch
def test_get_tensor():
    from deepchem.utils.differentiation_utils.editable_module import _get_tensors

    class A:

        def __init__(self):
            self.a = 2
            self.b = torch.tensor(3.0)
            self.c = torch.tensor(4.0)
            self.d = torch.tensor(5.0)

    a = A()
    outputs = _get_tensors(a)
    assert outputs[0] == [torch.tensor(3.), torch.tensor(4.), torch.tensor(5.)]
    assert outputs[1] == ['b', 'c', 'd']


@pytest.mark.torch
def test_set_tensor():
    from deepchem.utils.differentiation_utils.editable_module import _set_tensors

    class A:

        def __init__(self):
            self.a = 2
            self.b = torch.tensor(3.0)
            self.c = torch.tensor(4.0)
            self.d = torch.tensor(5.0)

    a = A()
    _set_tensors(a, [torch.tensor(6.), torch.tensor(7.), torch.tensor(8.)])
    assert a.b == torch.tensor(6.)
    assert a.c == torch.tensor(7.)


@pytest.mark.torch
def test_normalize_bcast_dims():
    from deepchem.utils.differentiation_utils import normalize_bcast_dims
    assert normalize_bcast_dims([1, 2, 3], [2, 3]) == [[1, 2, 3], [1, 2, 3]]


@pytest.mark.torch
def test_get_bcasted_dims():
    from deepchem.utils.differentiation_utils import get_bcasted_dims
    assert get_bcasted_dims([1, 2, 5], [2, 3, 4]) == [2, 3, 5]


@pytest.mark.torch
def test_match_dim():
    from deepchem.utils.differentiation_utils import match_dim
    x = torch.randn(10, 5)
    xq = torch.randn(10, 3)
    x_new, xq_new = match_dim(x, xq)
    assert x_new.shape == torch.Size([10, 5])
    assert xq_new.shape == torch.Size([10, 3])


@pytest.mark.torch
def test_linear_operator():
    from deepchem.utils.differentiation_utils import LinearOperator
    torch.manual_seed(100)

    class MyLinOp(LinearOperator):

        def __init__(self, shape):
            super(MyLinOp, self).__init__(shape)
            self.param = torch.rand(shape)

        def _getparamnames(self, prefix=""):
            return [prefix + "param"]

        def _mv(self, x):
            return torch.matmul(self.param, x)

        def _rmv(self, x):
            return torch.matmul(self.param.transpose(-2, -1).conj(), x)

        def _mm(self, x):
            return torch.matmul(self.param, x)

        def _rmm(self, x):
            return torch.matmul(self.param.transpose(-2, -1).conj(), x)

        def _fullmatrix(self):
            return self.param

    linop = MyLinOp((1, 3, 1, 2))
    x = torch.rand(1, 3, 2, 2)
    assert torch.allclose(linop.mv(x), torch.matmul(linop.param, x))
    x = torch.rand(1, 3, 1, 1)
    assert torch.allclose(linop.rmv(x),
                          torch.matmul(linop.param.transpose(-2, -1).conj(), x))
    x = torch.rand(1, 3, 2, 2)
    assert torch.allclose(linop.mm(x), torch.matmul(linop.param, x))
    x = torch.rand(1, 3, 1, 2)
    assert torch.allclose(linop.rmm(x),
                          torch.matmul(linop.param.transpose(-2, -1).conj(), x))
    assert torch.allclose(linop.fullmatrix(), linop.param)


@pytest.mark.torch
def test_add_linear_operator():
    from deepchem.utils.differentiation_utils import LinearOperator

    class Operator(LinearOperator):

        def __init__(self, mat: torch.Tensor, is_hermitian: bool) -> None:
            super(Operator, self).__init__(
                shape=mat.shape,
                is_hermitian=is_hermitian,
                dtype=mat.dtype,
                device=mat.device,
                _suppress_hermit_warning=True,
            )
            self.mat = mat

        def _mv(self, x: torch.Tensor) -> torch.Tensor:
            return torch.matmul(self.mat, x.unsqueeze(-1)).squeeze(-1)

        def _mm(self, x: torch.Tensor) -> torch.Tensor:
            return torch.matmul(self.mat, x)

        def _rmv(self, x: torch.Tensor) -> torch.Tensor:
            return torch.matmul(
                self.mat.transpose(-3, -1).conj(), x.unsqueeze(-1)).squeeze(-1)

        def _rmm(self, x: torch.Tensor) -> torch.Tensor:
            return torch.matmul(self.mat.transpose(-2, -1).conj(), x)

        def _fullmatrix(self) -> torch.Tensor:
            return self.mat

        def _getparamnames(self, prefix: str = ""):
            return [prefix + "mat"]

    op = Operator(torch.tensor([[1, 2.], [3, 4]]), is_hermitian=False)
    x = torch.tensor([[2, 2], [1, 2.]])
    op2 = op + op
    assert torch.allclose(op2.mm(x), 2 * op.mm(x))


@pytest.mark.torch
def test_mul_linear_operator():
    from deepchem.utils.differentiation_utils import LinearOperator

    class MyLinOp(LinearOperator):

        def __init__(self, shape):
            super(MyLinOp, self).__init__(shape)
            self.param = torch.rand(shape)

        def _getparamnames(self, prefix=""):
            return [prefix + "param"]

        def _mv(self, x):
            return torch.matmul(self.param, x)

    linop = MyLinOp((1, 3, 1, 2))
    linop2 = linop * 2
    x = torch.rand(1, 3, 2, 2)
    torch.allclose(linop.mv(x) * 2, linop2.mv(x))


@pytest.mark.torch
def test_adjoint_linear_operator():
    from deepchem.utils.differentiation_utils import LinearOperator

    class MyLinOp(LinearOperator):

        def __init__(self, shape):
            super(MyLinOp, self).__init__(shape)
            self.param = torch.rand(shape)

        def _getparamnames(self, prefix=""):
            return [prefix + "param"]

        def _mv(self, x):
            return torch.matmul(self.param, x)

        def _rmv(self, x):
            return torch.matmul(self.param.transpose(-2, -1).conj(), x)

    linop = MyLinOp((1, 3, 1, 2))
    x = torch.rand(1, 3, 1, 1)
    result_rmv = linop.rmv(x)

    adjoint_linop = linop.H
    result_mv = adjoint_linop.mv(x)

    assert torch.allclose(result_rmv, result_mv)


@pytest.mark.torch
def test_matmul_linear_operator():
    from deepchem.utils.differentiation_utils import LinearOperator

    class MyLinOp(LinearOperator):

        def __init__(self, shape):
            super(MyLinOp, self).__init__(shape)
            self.param = torch.rand(shape)

        def _getparamnames(self, prefix=""):
            return [prefix + "param"]

        def _mv(self, x):
            return torch.matmul(self.param, x)

    linop1 = MyLinOp((1, 3, 1, 2))
    linop2 = MyLinOp((1, 3, 2, 1))
    linop_result = linop1.matmul(linop2)
    x = torch.rand(1, 3, 1, 1)
    result = linop_result.mv(x)
    assert result.shape == torch.Size([1, 3, 1, 1])


@pytest.mark.torch
def test_matrix_linear_operator():
    from deepchem.utils.differentiation_utils import LinearOperator

    mat = torch.rand(2, 2)
    linop = LinearOperator.m(mat)
    x = torch.randn(2, 2)

    result_mm = linop.mm(x)
    expected_mm = torch.matmul(mat, x)

    result_mv = linop.mv(x)
    expected_mv = torch.matmul(mat, x.unsqueeze(-1)).squeeze(-1)

    assert torch.allclose(result_mm, expected_mm)
    assert torch.allclose(result_mv, expected_mv)


def test_set_default_options():
    from deepchem.utils.differentiation_utils import set_default_option
    assert set_default_option({'a': 1, 'b': 2}, {'a': 3}) == {'a': 3, 'b': 2}


def test_get_and_pop_keys():
    from deepchem.utils.differentiation_utils import get_and_pop_keys
    assert get_and_pop_keys({'a': 1, 'b': 2}, ['a']) == {'a': 1}


def test_get_method():
    from deepchem.utils.differentiation_utils import get_method
    assert get_method('foo', {'bar': lambda: 1}, 'bar')() == 1


def test_dummy_context_manager():
    """Just checks that dummy_context_manager doesn't crash"""
    from deepchem.utils.differentiation_utils import dummy_context_manager
    with dummy_context_manager() as x:
        if x is None:
            pass
        else:
            raise AssertionError()


def test_assert_runtime():
    from deepchem.utils.differentiation_utils import assert_runtime
    try:
        assert_runtime(False, "This should fail")
    except RuntimeError:
        pass



import numpy as np
from typing import List, Any
from numpy.typing import ArrayLike
from deepchem.feat.graph_data import BatchGraphData, GraphData

try:
    import torch
except ModuleNotFoundError:
    pass

from deepchem.feat.molecule_featurizers.dmpnn_featurizer import GraphConvConstants


class BatchGroverGraph:
    """Utility to batch graphs featurized by GroverFeaturizer

    GraphData created by created GroverFeaturizer has pre-computed attributes
    which needs to be aggregated before passing to the Grover model. This
    class batches graph by taking into account the pre-computed and additional
    attributes and returns a batched graph which can be passed to the Grover
    model.

    The batching method takes in a dc.feat.GraphData object created by
    GroverFeaturizer. New node indices and edge indices are
    computed by stacking the adjacency matrices of the graphs diagonally
    into a joint single adjacency matrix. On the new adjacency matrix,
    the method compute additional features like atom-to-bond indexes
    and bond-to-atom indexes. The rest of the features are computed by stacking
    the attributes as in dc.feat.BatchGraphData.


    Parameters
    ----------
    molgraph: List[GraphData]
        A list of GraphData objects created by GroverFeaturizer

    Example
    -------
    >>> import deepchem as dc
    >>> smiles = ['CC', 'CCC', 'CC(=O)C']
    >>> featurizer = dc.feat.GroverFeaturizer(features_generator=dc.feat.CircularFingerprint())
    >>> graphs = featurizer.featurize(smiles)
    >>> batched_graph = BatchGroverGraph(graphs)
    """

    def __init__(self, mol_graphs: List[GraphData]):
        self.smiles_batch = []
        self.n_mols = len(mol_graphs)

        self.atom_fdim = GraphConvConstants.ATOM_FDIM + 18
        self.bond_fdim = GraphConvConstants.BOND_FDIM + self.atom_fdim

        self.n_atoms = 0
        self.n_bonds = 0
        a_scope = [
        ]  # list of tuples indicating (start_atom_index, num_atoms) for each molecule
        b_scope = [
        ]  # list of tuples indicating (start_bond_index, num_bonds) for each molecule

        f_atoms: List[ArrayLike] = []  # atom features
        f_bonds: List[ArrayLike] = []  # combined atom/bond features
        a2b: Any = [[]]  # mapping from atom index to incoming bond indices
        b2a = [
        ]  # mapping from bond index to the index of the atom the bond is coming from
        b2revb = []  # mapping from bond index to the index of the reverse bond

        fg_labels = []
        additional_features = []

        # NOTE We have lot of type ignores here since grover mol-graph which is of type
        # GraphData have kwargs which are not attributes of GraphData. Hence, in these
        # cases mypy raises GraphData does not have attributes `..`.
        for mol_graph in mol_graphs:
            self.smiles_batch.append(mol_graph.smiles)  # type: ignore
            fg_labels.append(mol_graph.fg_labels)  # type: ignore
            additional_features.append(
                mol_graph.additional_features)  # type: ignore
            f_atoms.extend(mol_graph.node_features)
            f_bonds.extend(mol_graph.edge_features)  # type: ignore

            for a in range(mol_graph.n_atoms):  # type: ignore
                a2b.append([
                    b + self.n_bonds for b in mol_graph.a2b[a]  # type: ignore
                ])  # type: ignore

            for b in range(mol_graph.n_bonds):  # type: ignore
                b2a.append(self.n_atoms + mol_graph.b2a[b])  # type: ignore
                b2revb.append(self.n_bonds +
                              mol_graph.b2revb[b])  # type: ignore

            a_scope.append((self.n_atoms, mol_graph.n_atoms))  # type: ignore
            b_scope.append((self.n_bonds, mol_graph.n_bonds))  # type: ignore
            self.n_atoms += mol_graph.n_atoms  # type: ignore
            self.n_bonds += mol_graph.n_bonds  # type: ignore

        # max with 1 to fix a crash in rare case of all single-heavy-atom mols
        self.max_num_bonds = max(1, max(len(in_bonds) for in_bonds in a2b))

        self.f_atoms = torch.FloatTensor(np.asarray(f_atoms))
        self.f_bonds = torch.FloatTensor(np.asarray(f_bonds))
        self.a2b = torch.LongTensor(
            np.asarray([
                a2b[a] + [0] * (self.max_num_bonds - len(a2b[a]))
                for a in range(self.n_atoms)
            ]))
        self.b2a = torch.LongTensor(np.asarray(b2a))
        self.b2revb = torch.LongTensor(np.asarray(b2revb))
        self.a2a = self.b2a[self.a2b]  # only needed if using atom messages
        self.a_scope = torch.LongTensor(a_scope)
        self.b_scope = torch.LongTensor(b_scope)

        self.fg_labels = torch.Tensor(np.asarray(fg_labels)).float()
        self.additional_features = torch.from_numpy(
            np.stack(additional_features)).float()

    def get_components(self):
        """Returns the components of BatchGroverGraph.

        Example
        -------
        >>> import deepchem as dc
        >>> smiles = ['CC', 'CCC', 'CC(=O)C']
        >>> featurizer = dc.feat.GroverFeaturizer(features_generator=dc.feat.CircularFingerprint())
        >>> graphs = featurizer.featurize(smiles)
        >>> batched_graph = BatchGroverGraph(graphs)
        >>> components = batched_graph.get_components()

        Returns
        -------
        components: Tuple
            A tuple containing PyTorch tensors with the atom features, bond features, graph structure,
            two lists indicating the scope of the atoms and bonds (i.e. which molecules they belong to)
            and functional group labels.
        """
        return self.f_atoms, self.f_bonds, self.a2b, self.b2a, self.b2revb, self.a2a, self.a_scope, self.b_scope, self.fg_labels


def _get_atom_scopes(graph_index: ArrayLike) -> List[List[int]]:
    """Atom scope is a list of tuples with a single entry for every
    molecule in the batched graph. The entry indicates the beginning
    node index for a molecule and the number of nodes in the molecule.

    Parameters
    ----------
    graph_index: np.array
        An array containing a mapping between node index and the graph
    in the batched graph.

    Returns
    -------
    scopes: List[List[int]]
        Node index scope for each molecule in the batched graph.

    Example
    -------
    >>> import numpy as np
    >>> graph_index = np.array([0, 0, 1, 1, 1])
    >>> _get_atom_scopes(graph_index)
    [[0, 2], [2, 3]]
    """
    # graph_index indicates which atom belongs to which molecule
    mols = np.unique(graph_index)
    scopes = []
    for mol in mols:
        positions = np.where(graph_index == mol, 1, 0)
        scopes.append(
            [int(np.argmax(positions)),
             int(np.count_nonzero(positions))])
    return scopes


def _get_bond_scopes(edge_index: ArrayLike,
                     graph_index: ArrayLike) -> List[List[int]]:
    """Bond scope is a list of tuples with a single entry for every molecule
    in the batched graph. The entry indicates the beginning bond index for a
    molecule and the number of bonds in the molecule.

    Parameters
    ----------
    edge_index: np.array
        Graph connectivity in COO format with shape [2, num_edges]
    graph_index: np.array
        An array containing a mapping between node index and the graph
    in the batched graph.

    Returns
    -------
    scopes: List[List[int]]
        Bond index scope for each molecule in the batched graph.

    Example
    -------
    >>> edge_index = np.array([[0, 1, 2, 4], [1, 0, 4, 2]]) # a molecule with 4 bonds
    >>> graph_index = np.array([0, 0, 1, 1, 1])
    >>> _get_bond_scopes(edge_index, graph_index)
    [[0, 2], [2, 2]]
    """
    mols = np.unique(graph_index)
    bond_index = graph_index[edge_index[0]]  # type: ignore
    scopes = []
    for mol in mols:
        positions = np.where(bond_index == mol, 1, 0)
        scopes.append(
            [int(np.argmax(positions)),
             int(np.count_nonzero(positions))])
    return scopes


def _compute_b2revb(edge_index: np.ndarray) -> List[int]:
    """Every edge in a grover graph is a directed edge. Hence, a bond
    is represented by two edges of opposite directions. b2revb is a representation
    which stores for every edge, the index of reverse edge of that edge.

    Parameters
    ----------
    edge_index: np.array
        Graph connectivity in COO format with shape [2, num_edges]

    Returns
    -------
    b2revb: List[int]
        A mapping where an element at an index contains the index of the reverse bond.

    Example
    -------
    >>> import numpy as np
    >>> edge_index = np.array([[0, 1, 2, 4], [1, 0, 4, 2]])
    >>> _compute_b2revb(edge_index)
    [1, 0, 3, 2]
    """
    b2revb = [0] * edge_index.shape[1]
    for i, bond in enumerate(edge_index.T):
        for j, (sa, da) in enumerate(edge_index.T):
            if sa == bond[1] and da == bond[0]:
                b2revb[i] = j
    return b2revb


def _get_a2b(n_atoms: int, edge_index: np.ndarray) -> np.ndarray:
    """a2b is a mapping between atoms and their incoming bonds.

    Parameters
    ----------
    n_atoms: int
        Number of atoms
    edge_index: np.array
        Graph connectivity in COO format with shape [2, num_edges]

    Returns
    -------
    a2b: ArrayLike
        A mapping between atoms and their incoming bonds

    Example
    -------
    >>> import numpy as np
    >>> edge_index = np.array([[0, 1], [1, 2]])
    >>> n_atoms = 3
    >>> _get_a2b(n_atoms, edge_index)
    array([[0],
           [0],
           [1]])
    """
    a2b: List[List[Any]] = [[] for atom in range(n_atoms)]

    for i, bond in enumerate(edge_index.T):
        dest_atom = bond[1]
        a2b[dest_atom].append(i)

    # padding
    max_num_bonds = max(map(lambda x: len(x), a2b))
    atom_bond_mapping = np.asarray(
        [a2b[a] + [0] * (max_num_bonds - len(a2b[a])) for a in range(n_atoms)])

    return atom_bond_mapping


def extract_grover_attributes(molgraph: BatchGraphData):
    """Utility to extract grover attributes for grover model

    Parameters
    ----------
    molgraph: BatchGraphData
        A batched graph data representing a collection of molecules.

    Returns
    -------
    graph_attributes: Tuple
        A tuple containing atom features, bond features, atom to bond mapping, bond to atom mapping, bond to reverse bond mapping, atom to atom mapping, atom scope, bond scope, functional group labels and other additional features.

    Example
    -------
    >>> import deepchem as dc
    >>> from deepchem.feat.graph_data import BatchGraphData
    >>> smiles = ['CC', 'CCC', 'CC(=O)C']
    >>> featurizer = dc.feat.GroverFeaturizer(features_generator=dc.feat.CircularFingerprint())
    >>> graphs = featurizer.featurize(smiles)
    >>> molgraph = BatchGraphData(graphs)
    >>> attributes = extract_grover_attributes(molgraph)
    """
    fg_labels = getattr(molgraph, 'fg_labels')
    additional_features = getattr(molgraph, 'additional_features')
    f_atoms = molgraph.node_features
    f_bonds = molgraph.edge_features
    graph_index = molgraph.graph_index
    edge_index = molgraph.edge_index

    a_scope = _get_atom_scopes(graph_index)
    b_scope = _get_bond_scopes(edge_index, graph_index)
    b2revb = _compute_b2revb(edge_index)

    # computing a2b
    a2b = _get_a2b(molgraph.num_nodes, edge_index)

    f_atoms_tensor = torch.FloatTensor(f_atoms)
    f_bonds_tensor = torch.FloatTensor(f_bonds)
    fg_labels_tensor = torch.FloatTensor(fg_labels)
    additional_features_tensor = torch.FloatTensor(additional_features)
    a2b_tensor = torch.LongTensor(a2b)
    b2a_tensor = torch.LongTensor(molgraph.edge_index[0])
    b2revb_tensor = torch.LongTensor(b2revb)
    # only needed if using atom messages
    a2a = b2a_tensor[a2b_tensor]  # type: ignore
    a_scope_tensor = torch.LongTensor(np.asarray(a_scope))
    b_scope_tensor = torch.LongTensor(np.asarray(b_scope))
    return f_atoms_tensor, f_bonds_tensor, a2b_tensor, b2a_tensor, b2revb_tensor, a2a, a_scope_tensor, b_scope_tensor, fg_labels_tensor, additional_features_tensor



import torch
import numpy as np
from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.models.torch_models.layers import AtomicConv
from deepchem.data import Dataset
from deepchem.models.losses import L2Loss

from typing import List, Callable, Optional, Sequence, Tuple, Iterable
from deepchem.utils.typing import OneOrMany, ActivationFn


class AtomConvModel(TorchModel):
    """An Atomic Convolutional Neural Network (ACNN) for energy score prediction.

    The network follows the design of a graph convolutional network but in this case the graph is represented
    as a 3D structure of the molecule. The objective of this model is to train models and predict energetic
    state starting from the spatial geometry of the model [1].

    References
    ----------
    .. [1] Gomes, Joseph, et al. "Atomic convolutional networks for predicting protein-ligand binding affinity." arXiv preprint arXiv:1703.10603 (2017).

    Examples
    --------
    >>> from deepchem.models.torch_models import AtomConvModel
    >>> frag1_num_atoms = 100 # atoms for ligand
    >>> frag2_num_atoms = 1200 # atoms for protein
    >>> complex_num_atoms = frag1_num_atoms + frag2_num_atoms
    >>> batch_size = 1
    >>> # Initialize the model
    >>> atomic_convnet = AtomConvModel(n_tasks=1,
                                   batch_size=batch_size,
                                   layer_sizes=[
                                       10,
                                   ],
                                   frag1_num_atoms=frag1_num_atoms,
                                   frag2_num_atoms=frag2_num_atoms,
                                   complex_num_atoms=complex_num_atoms)

    # Creates a set of dummy features that contain the coordinate and
    # neighbor-list features required by the AtomicConvModel.
    >>> # Preparing the dataset
    >>> features = []
    >>> frag1_coords = np.random.rand(frag1_num_atoms, 3)
    >>> frag1_nbr_list = {i: [] for i in range(frag1_num_atoms)}
    >>> frag1_z = np.random.randint(10, size=(frag1_num_atoms))
    >>> frag2_coords = np.random.rand(frag2_num_atoms, 3)
    >>> frag2_nbr_list = {i: [] for i in range(frag2_num_atoms)}
    >>> frag2_z = np.random.randint(10, size=(frag2_num_atoms))
    >>> system_coords = np.random.rand(complex_num_atoms, 3)
    >>> system_nbr_list = {i: [] for i in range(complex_num_atoms)}
    >>> system_z = np.random.randint(10, size=(complex_num_atoms))
    >>> features.append((frag1_coords, frag1_nbr_list, frag1_z, frag2_coords, frag2_nbr_list, frag2_z, system_coords, system_nbr_list, system_z))
    >>> features = np.asarray(features)
    >>> labels = np.zeros(batch_size)
    >>> train = NumpyDataset(features, labels)
    >>> atomic_convnet.fit(train, nb_epoch=1)
    >>> preds = atomic_convnet.predict(train)
    """

    def __init__(self,
                 n_tasks: int,
                 frag1_num_atoms: int = 70,
                 frag2_num_atoms: int = 634,
                 complex_num_atoms: int = 701,
                 max_num_neighbors: int = 12,
                 batch_size: int = 24,
                 atom_types: Sequence[float] = [
                     6, 7., 8., 9., 11., 12., 15., 16., 17., 20., 25., 30., 35.,
                     53., -1.
                 ],
                 radial: Sequence[Sequence[float]] = [[
                     1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0, 5.5, 6.0, 6.5, 7.0,
                     7.5, 8.0, 8.5, 9.0, 9.5, 10.0, 10.5, 11.0, 11.5, 12.0
                 ], [0.0, 4.0, 8.0], [0.4]],
                 layer_sizes=[100],
                 weight_init_stddevs: OneOrMany[float] = 0.02,
                 bias_init_consts: OneOrMany[float] = 1.0,
                 weight_decay_penalty: float = 0.0,
                 weight_decay_penalty_type: str = "l2",
                 dropouts: OneOrMany[float] = 0.5,
                 activation_fns: OneOrMany[ActivationFn] = ['relu'],
                 residual: bool = False,
                 learning_rate=0.001,
                 **kwargs) -> None:
        """TorchModel wrapper for ACNN

         Parameters
        ----------
        n_tasks: int
            number of tasks
        frag1_num_atoms: int
            Number of atoms in first fragment.
        frag2_num_atoms: int
            Number of atoms in second fragment.
        complex_num_atoms: int
            Number of atoms in complex.
        max_num_neighbors: int
            Maximum number of neighbors possible for an atom. Recall neighbors
            are spatial neighbors.
        batch_size: int
            Size of the batch.
        atom_types: list
            List of atoms recognized by model. Atoms are indicated by their
            nuclear numbers.
        radial: list
            Radial parameters used in the atomic convolution transformation.
        layer_sizes: list
            the size of each dense layer in the network.  The length of
            this list determines the number of layers.
        weight_init_stddevs: list or float
            the standard deviation of the distribution to use for weight
            initialization of each layer.  The length of this list should
            equal len(layer_sizes).  Alternatively, this may be a single
            value instead of a list, where the same value is used
            for every layer.
        bias_init_consts: list or float
            the value to initialize the biases in each layer.  The
            length of this list should equal len(layer_sizes).
            Alternatively, this may be a single value instead of a list, where the same value is used for every layer.
        dropouts: list or float
            the dropout probability to use for each layer.  The length of this list should equal len(layer_sizes).
            Alternatively, this may be a single value instead of a list, where the same value is used for every layer.
        activation_fns: list or object
            the Tensorflow activation function to apply to each layer.  The length of this list should equal
            len(layer_sizes).  Alternatively, this may be a single value instead of a list, where the
            same value is used for every layer.
        residual:  bool
            Whether to use residual connections.
        learning_rate: float
            the learning rate to use for fitting.
        """

        self.n_tasks = n_tasks
        self.complex_num_atoms = complex_num_atoms
        self.frag1_num_atoms = frag1_num_atoms
        self.frag2_num_atoms = frag2_num_atoms
        self.max_num_neighbors = max_num_neighbors
        self.batch_size = batch_size
        self.atom_types = atom_types

        self.model = AtomicConv(n_tasks=n_tasks,
                                frag1_num_atoms=frag1_num_atoms,
                                frag2_num_atoms=frag2_num_atoms,
                                complex_num_atoms=complex_num_atoms,
                                max_num_neighbors=max_num_neighbors,
                                batch_size=batch_size,
                                atom_types=atom_types,
                                radial=radial,
                                layer_sizes=layer_sizes,
                                weight_init_stddevs=weight_init_stddevs,
                                bias_init_consts=bias_init_consts,
                                dropouts=dropouts,
                                activation_fns=activation_fns,
                                residual=residual,
                                learning_rate=learning_rate)

        regularization_loss: Optional[Callable]

        if weight_decay_penalty != 0:
            weights = [layer.weight for layer in self.model.layers]
            if weight_decay_penalty_type == 'l1':
                regularization_loss = lambda: weight_decay_penalty * torch.sum(  # noqa: E731
                    torch.stack([torch.abs(w).sum() for w in weights]))
            else:
                regularization_loss = lambda: weight_decay_penalty * torch.sum(  # noqa: E731
                    torch.stack([torch.square(w).sum() for w in weights]))
        else:
            regularization_loss = None

        loss = L2Loss()

        super(AtomConvModel,
              self).__init__(self.model,
                             loss=loss,
                             batch_size=batch_size,
                             regularization_loss=regularization_loss,
                             **kwargs)

    def default_generator(
            self,
            dataset: Dataset,
            epochs: int = 1,
            mode: str = 'fit',
            deterministic: bool = True,
            pad_batches: bool = True) -> Iterable[Tuple[List, List, List]]:
        """Convert a dataset into the tensors needed for learning.

        Parameters
        ----------
        dataset: `dc.data.Dataset`
            Dataset to convert
        epochs: int, optional (Default 1)
            Number of times to walk over `dataset`
        mode: str, optional (Default 'fit')
            Ignored in this implementation.
        deterministic: bool, optional (Default True)
            Whether the dataset should be walked in a deterministic fashion
        pad_batches: bool, optional (Default True)
            If true, each returned batch will have size `self.batch_size`.

        Returns
        -------
        Iterator which walks over the batches
        """

        batch_size = self.batch_size

        def replace_atom_types(z):
            """Replace the atom types depending by their nuclear numbers to "-1" value if reapeated for the training loop.

            Parameters
            ----------
            z: list
                Atom types learned from the model.

            Returns
            -------
            A list of nuclear numbers with "-1" values in the repeated indexes of inverted z.
            """
            np.putmask(z, np.isin(z, list(self.atom_types), invert=True), -1)
            return z

        for epoch in range(epochs):
            for ind, (F_b, y_b, w_b, ids_b) in enumerate(
                    dataset.iterbatches(batch_size,
                                        deterministic=True,
                                        pad_batches=pad_batches)):

                N = self.complex_num_atoms
                N_1 = self.frag1_num_atoms
                N_2 = self.frag2_num_atoms
                M = self.max_num_neighbors

                batch_size = F_b.shape[0]
                num_features = F_b[0][0].shape[1]
                frag1_X_b = np.zeros((batch_size, N_1, num_features))
                for i in range(batch_size):
                    frag1_X_b[i] = F_b[i][0]

                frag2_X_b = np.zeros((batch_size, N_2, num_features))
                for i in range(batch_size):
                    frag2_X_b[i] = F_b[i][3]

                complex_X_b = np.zeros((batch_size, N, num_features))
                for i in range(batch_size):
                    complex_X_b[i] = F_b[i][6]

                frag1_Nbrs = np.zeros((batch_size, N_1, M))
                frag1_Z_b = np.zeros((batch_size, N_1))
                for i in range(batch_size):
                    z = replace_atom_types(F_b[i][2])
                    frag1_Z_b[i] = z
                frag1_Nbrs_Z = np.zeros((batch_size, N_1, M))
                for atom in range(N_1):
                    for i in range(batch_size):
                        atom_nbrs = F_b[i][1].get(atom, "")
                        frag1_Nbrs[i,
                                   atom, :len(atom_nbrs)] = np.array(atom_nbrs)
                        for j, atom_j in enumerate(atom_nbrs):
                            frag1_Nbrs_Z[i, atom, j] = frag1_Z_b[i, atom_j]

                frag2_Nbrs = np.zeros((batch_size, N_2, M))
                frag2_Z_b = np.zeros((batch_size, N_2))
                for i in range(batch_size):
                    z = replace_atom_types(F_b[i][5])
                    frag2_Z_b[i] = z
                frag2_Nbrs_Z = np.zeros((batch_size, N_2, M))
                for atom in range(N_2):
                    for i in range(batch_size):
                        atom_nbrs = F_b[i][4].get(atom, "")
                        frag2_Nbrs[i,
                                   atom, :len(atom_nbrs)] = np.array(atom_nbrs)
                        for j, atom_j in enumerate(atom_nbrs):
                            frag2_Nbrs_Z[i, atom, j] = frag2_Z_b[i, atom_j]

                complex_Nbrs = np.zeros((batch_size, N, M))
                complex_Z_b = np.zeros((batch_size, N))
                for i in range(batch_size):
                    z = replace_atom_types(F_b[i][8])
                    complex_Z_b[i] = z
                complex_Nbrs_Z = np.zeros((batch_size, N, M))
                for atom in range(N):
                    for i in range(batch_size):
                        atom_nbrs = F_b[i][7].get(atom, "")
                        complex_Nbrs[i, atom, :len(atom_nbrs)] = np.array(
                            atom_nbrs)
                        for j, atom_j in enumerate(atom_nbrs):
                            complex_Nbrs_Z[i, atom, j] = complex_Z_b[i, atom_j]

                inputs = [
                    frag1_X_b, frag1_Nbrs, frag1_Nbrs_Z, frag1_Z_b, frag2_X_b,
                    frag2_Nbrs, frag2_Nbrs_Z, frag2_Z_b, complex_X_b,
                    complex_Nbrs, complex_Nbrs_Z, complex_Z_b
                ]

                y_b = np.reshape(y_b, newshape=(batch_size, 1))

                yield (inputs, [y_b], [w_b])



"""
Utilities for miscellaneous tasks.
"""


def indent(s, nspace):
    """Gives indentation of the second line and next lines.
    It is used to format the string representation of an object.
    Which might be containing multiples objects in it.
    Usage: LinearOperator

    Parameters
    ----------
    s: str
        The string to be indented.
    nspace: int
        The number of spaces to be indented.

    Returns
    -------
    str
        The indented string.

    """
    spaces = " " * nspace
    lines = [spaces + c if i > 0 else c for i, c in enumerate(s.split("\n"))]
    return "\n".join(lines)


def shape2str(shape):
    """Convert the shape to string representation.
    It also nicely formats the shape to be readable.

    Parameters
    ----------
    shape: Sequence[int]
        The shape to be converted to string representation.

    Returns
    -------
    str
        The string representation of the shape.

    """
    return "(%s)" % (", ".join([str(s) for s in shape]))



import logging
try:
    import jax.numpy as jnp
    import haiku as hk
except:
    raise ImportError('These classes require Jax and Haiku to be installed.')

logger = logging.getLogger(__name__)


class Linear(hk.Module):
    """Protein folding specific Linear Module.

    This differs from the standard Haiku Linear in a few ways:
        * It supports inputs of arbitrary rank
        * Initializers are specified by strings

    This code is adapted from DeepMind's AlphaFold code release
    (https://github.com/deepmind/alphafold).

    Examples
    --------
    >>> import deepchem as dc
    >>> import haiku as hk
    >>> import jax
    >>> import deepchem.models.jax_models.layers
    >>> def forward_model(x):
    ...   layer = dc.models.jax_models.layers.Linear(2)
    ...   return layer(x)
    >>> f = hk.transform(forward_model)
    >>> rng = jax.random.PRNGKey(42)
    >>> x = jnp.ones([8, 28 * 28])
    >>> params = f.init(rng, x)
    >>> output = f.apply(params, rng, x)
    """

    def __init__(self,
                 num_output: int,
                 initializer: str = 'linear',
                 use_bias: bool = True,
                 bias_init: float = 0.,
                 name: str = 'linear'):
        """Constructs Linear Module.

        Parameters
        ----------
        num_output: int
            number of output channels.
        initializer: str (default 'linear')
            What initializer to use, should be one of {'linear', 'relu', 'zeros'}
        use_bias: bool (default True)
            Whether to include trainable bias
        bias_init: float (default 0)
            Value used to initialize bias.
        name: str (default 'linear')
            name of module, used for name scopes.
        """

        super().__init__(name=name)
        self.num_output = num_output
        self.initializer = initializer
        self.use_bias = use_bias
        self.bias_init = bias_init

    def __call__(self, inputs: jnp.ndarray) -> jnp.ndarray:
        """Connects Module.

        Parameters
        ----------
        inputs: jnp.ndarray
            Tensor of shape [..., num_channel]

        Returns
        -------
        output of shape [..., num_output]
        """
        n_channels = int(inputs.shape[-1])

        weight_shape = [n_channels, self.num_output]
        if self.initializer == 'linear':
            weight_init = hk.initializers.VarianceScaling(mode='fan_in',
                                                          scale=1.)
        elif self.initializer == 'relu':
            weight_init = hk.initializers.VarianceScaling(mode='fan_in',
                                                          scale=2.)
        elif self.initializer == 'zeros':
            weight_init = hk.initializers.Constant(0.0)

        weights = hk.get_parameter('weights', weight_shape, inputs.dtype,
                                   weight_init)

        # this is equivalent to einsum('...c,cd->...d', inputs, weights)
        # but turns out to be slightly faster
        inputs = jnp.swapaxes(inputs, -1, -2)
        output = jnp.einsum('...cb,cd->...db', inputs, weights)
        output = jnp.swapaxes(output, -1, -2)

        if self.use_bias:
            bias = hk.get_parameter('bias', [self.num_output], inputs.dtype,
                                    hk.initializers.Constant(self.bias_init))
            output += bias

        return output



"""Model-Agnostic Meta-Learning (MAML) algorithm for low data learning."""

import os
import shutil
import tempfile
import time

from deepchem.models.optimizers import Optimizer, Adam, GradientDescent, LearningRateSchedule
from typing import Any, Dict, List, Optional, Tuple, Sequence, Union
from deepchem.utils.typing import OneOrMany

try:
    from deepchem.metalearning import MetaLearner
    import torch
    has_pytorch = True
except:
    has_pytorch = False


class MAML(object):
    """Implements the Model-Agnostic Meta-Learning algorithm for low data learning.

    The algorithm is described in Finn et al., "Model-Agnostic Meta-Learning for Fast
    Adaptation of Deep Networks" (https://arxiv.org/abs/1703.03400).  It is used for
    training models that can perform a variety of tasks, depending on what data they
    are trained on.  It assumes you have training data for many tasks, but only a small
    amount for each one.  It performs "meta-learning" by looping over tasks and trying
    to minimize the loss on each one *after* one or a few steps of gradient descent.
    That is, it does not try to create a model that can directly solve the tasks, but
    rather tries to create a model that is very easy to train.

    To use this class, create a subclass of MetaLearner that encapsulates the model
    and data for your learning problem.  Pass it to a MAML object and call fit().
    You can then use train_on_current_task() to fine tune the model for a particular
    task.
    Example
    --------
    >>> import deepchem as dc
    >>> import numpy as np
    >>> import torch
    >>> import torch.nn.functional as F
    >>> from deepchem.metalearning.torch_maml import MetaLearner, MAML
    >>> class SineLearner(MetaLearner):
    ...     def __init__(self):
    ...         self.batch_size = 10
    ...         self.w1 = torch.nn.Parameter(torch.tensor(np.random.normal(size=[1, 40], scale=1.0),requires_grad=True))
    ...         self.w2 = torch.nn.Parameter(torch.tensor(np.random.normal(size=[40, 40], scale=np.sqrt(1 / 40)),requires_grad=True))
    ...         self.w3 = torch.nn.Parameter(torch.tensor(np.random.normal(size=[40, 1], scale=np.sqrt(1 / 40)),requires_grad=True))
    ...         self.b1 = torch.nn.Parameter(torch.tensor(np.zeros(40)),requires_grad=True)
    ...         self.b2 = torch.nn.Parameter(torch.tensor(np.zeros(40)),requires_grad=True)
    ...         self.b3 = torch.nn.Parameter(torch.tensor(np.zeros(1)),requires_grad=True)
    ...     def compute_model(self, inputs, variables, training):
    ...         x, y = inputs
    ...         w1, w2, w3, b1, b2, b3 = variables
    ...         dense1 = F.relu(torch.matmul(x, w1) + b1)
    ...         dense2 = F.relu(torch.matmul(dense1, w2) + b2)
    ...         output = torch.matmul(dense2, w3) + b3
    ...         loss = torch.mean(torch.square(output - y))
    ...         return loss, [output]
    ...     @property
    ...     def variables(self):
    ...         return [self.w1, self.w2, self.w3, self.b1, self.b2, self.b3]
    ...     def select_task(self):
    ...         self.amplitude = 5.0 * np.random.random()
    ...         self.phase = np.pi * np.random.random()
    ...     def get_batch(self):
    ...         x = torch.tensor(np.random.uniform(-5.0, 5.0, (self.batch_size, 1)))
    ...         return [x, torch.tensor(self.amplitude * np.sin(x + self.phase))]
    ...     def parameters(self):
    ...         for key, value in self.__dict__.items():
    ...             if isinstance(value, torch.nn.Parameter):
    ...                 yield value
    >>> learner = SineLearner()
    >>> optimizer = dc.models.optimizers.Adam(learning_rate=5e-3)
    >>> maml = MAML(learner,meta_batch_size=4,optimizer=optimizer)
    >>> maml.fit(9000)

    To test it out on a new task and see how it works

    >>> learner.select_task()
    >>> maml.restore()
    >>> batch = learner.get_batch()
    >>> loss, outputs = maml.predict_on_batch(batch)
    >>> maml.train_on_current_task()
    >>> loss, outputs = maml.predict_on_batch(batch)
    """

    def __init__(
        self,
        learner: MetaLearner,
        learning_rate: Union[float, LearningRateSchedule] = 0.001,
        optimization_steps: int = 1,
        meta_batch_size: int = 10,
        optimizer: Optimizer = Adam(),
        model_dir: Optional[str] = None,
        device: Optional[torch.device] = None,
    ):
        """Create an object for performing meta-optimization.

        Parameters
        ----------
        learner: MetaLearner
            defines the meta-learning problem
        learning_rate: float or Tensor
            the learning rate to use for optimizing each task (not to be confused with the one used
            for meta-learning).  This can optionally be made a variable (represented as a
            Tensor), in which case the learning rate will itself be learnable.
        optimization_steps: int
            the number of steps of gradient descent to perform for each task
        meta_batch_size: int
            the number of tasks to use for each step of meta-learning
        optimizer: Optimizer
            the optimizer to use for meta-learning (not to be confused with the gradient descent
            optimization performed for each task)
        model_dir: str
            the directory in which the model will be saved.  If None, a temporary directory will be created.
        device: torch.device, optional (default None)
            the device on which to run computations.  If None, a device is
            chosen automatically.
        """
        # Record inputs.

        self.learner: MetaLearner = learner
        self.learning_rate: Union[float, LearningRateSchedule] = learning_rate
        self.optimization_steps: int = optimization_steps
        self.meta_batch_size: int = meta_batch_size
        self.optimizer: Optimizer = optimizer

        # Create the output directory if necessary.

        self._model_dir_is_temp: bool = False
        if model_dir is not None:
            if not os.path.exists(model_dir):
                os.makedirs(model_dir)
        else:
            model_dir = tempfile.mkdtemp()
            self._model_dir_is_temp = True
        self.model_dir = model_dir
        self.save_file: str = "%s/%s" % (self.model_dir, "model")

        # Select a device.

        if device is None:
            if torch.cuda.is_available():
                device = torch.device('cuda')
            elif torch.backends.mps.is_available():
                device = torch.device('mps')
            else:
                device = torch.device('cpu')
        self.device: torch.device = device
        for param in learner.parameters():
            param = param.to(device)
            param.requires_grad_()

        # Create the optimizers for meta-optimization and task optimization.

        self._global_step: int = 0
        self._pytorch_optimizer = self.optimizer._create_pytorch_optimizer(
            learner.parameters())
        if isinstance(self.optimizer.learning_rate, LearningRateSchedule):
            self._lr_schedule = self.optimizer.learning_rate._create_pytorch_schedule(
                self._pytorch_optimizer)
        else:
            self._lr_schedule = None

        task_optimizer: Optimizer = GradientDescent(
            learning_rate=self.learning_rate)
        self._pytorch_task_optimizer = task_optimizer._create_pytorch_optimizer(
            learner.parameters())
        if isinstance(task_optimizer.learning_rate, LearningRateSchedule):
            self._lr_schedule = task_optimizer.learning_rate._create_pytorch_schedule(
                self._pytorch_task_optimizer)
        else:
            self._lr_schedule = None

    def __del__(self):
        if '_model_dir_is_temp' in dir(self) and self._model_dir_is_temp:
            shutil.rmtree(self.model_dir)

    def fit(self,
            steps: int,
            max_checkpoints_to_keep: int = 5,
            checkpoint_interval: int = 600,
            restore: bool = False):
        """Perform meta-learning to train the model.

        Parameters
        ----------
        steps: int
            the number of steps of meta-learning to perform
        max_checkpoints_to_keep: int
            the maximum number of checkpoint files to keep.  When this number is reached, older
            files are deleted.
        checkpoint_interval: int
            the time interval at which to save checkpoints, measured in seconds
        restore: bool
            if True, restore the model from the most recent checkpoint before training
            it further
        """
        if restore:
            self.restore()
        checkpoint_time: float = time.time()

        # Main optimization loop.

        learner = self.learner
        variables: OneOrMany[torch.Tensor] = learner.variables
        for i in range(steps):
            self._pytorch_optimizer.zero_grad()
            for j in range(self.meta_batch_size):
                learner.select_task()
                updated_variables: OneOrMany[torch.Tensor] = variables
                for k in range(self.optimization_steps):
                    loss, _ = self.learner.compute_model(
                        learner.get_batch(), updated_variables, True)
                    gradients: Tuple[torch.Tensor, ...] = torch.autograd.grad(
                        loss,
                        updated_variables,
                        grad_outputs=torch.ones_like(loss),
                        create_graph=True,
                        retain_graph=True)
                    updated_variables = [
                        v if g is None else v - self.learning_rate * g
                        for v, g in zip(updated_variables, gradients)
                    ]
                meta_loss, _ = self.learner.compute_model(
                    learner.get_batch(), updated_variables, True)
                meta_gradients: Tuple[torch.Tensor, ...] = torch.autograd.grad(
                    meta_loss,
                    variables,
                    grad_outputs=torch.ones_like(meta_loss),
                    retain_graph=True)
                if j == 0:
                    summed_gradients: Union[Tuple[torch.Tensor, ...],
                                            List[torch.Tensor]] = meta_gradients
                else:
                    summed_gradients = [
                        s + g for s, g in zip(summed_gradients, meta_gradients)
                    ]
                ind: int = 0
                for param in self.learner.parameters():
                    param.grad = summed_gradients[ind]
                    ind = ind + 1

            self._pytorch_optimizer.step()
            if self._lr_schedule is not None:
                self._lr_schedule.step()

            # Do checkpointing.

            if i == steps - 1 or time.time(
            ) >= checkpoint_time + checkpoint_interval:
                self.save_checkpoint(max_checkpoints_to_keep)
                checkpoint_time = time.time()

    def restore(self) -> None:
        """Reload the model parameters from the most recent checkpoint file."""
        last_checkpoint: Union[List[str], str] = sorted(
            self.get_checkpoints(self.model_dir))
        if len(last_checkpoint) == 0:
            raise ValueError('No checkpoint found')
        last_checkpoint = last_checkpoint[0]
        data: Any = torch.load(last_checkpoint, map_location=self.device)
        self.learner.__dict__ = data['model_state_dict']
        self.learning_rate = data['learning_rate']
        self._pytorch_optimizer.load_state_dict(data['optimizer_state_dict'])
        self._pytorch_task_optimizer.load_state_dict(
            data['task_optimizer_state_dict'])
        self._global_step = data['global_step']

    def train_on_current_task(self,
                              optimization_steps: int = 1,
                              restore: bool = True):
        """Perform a few steps of gradient descent to fine tune the model on the current task.

        Parameters
        ----------
        optimization_steps: int
            the number of steps of gradient descent to perform
        restore: bool
            if True, restore the model from the most recent checkpoint before optimizing
        """
        if restore:
            self.restore()
        variables: OneOrMany[torch.Tensor] = self.learner.variables
        task_optimizer: Optimizer = GradientDescent(
            learning_rate=self.learning_rate)
        self._pytorch_task_optimizer = task_optimizer._create_pytorch_optimizer(
            self.learner.parameters())
        if isinstance(task_optimizer.learning_rate, LearningRateSchedule):
            self._lr_schedule = task_optimizer.learning_rate._create_pytorch_schedule(
                self._pytorch_task_optimizer)
        else:
            self._lr_schedule = None
        for i in range(optimization_steps):
            self._pytorch_task_optimizer.zero_grad()
            inputs = self.learner.get_batch()
            loss, _ = self.learner.compute_model(inputs, variables, True)
            loss.backward()
            self._pytorch_task_optimizer.step()

    def predict_on_batch(
        self, inputs: OneOrMany[torch.Tensor]
    ) -> Tuple[torch.Tensor, Sequence[torch.Tensor]]:
        """Compute the model's outputs for a batch of inputs.

        Parameters
        ----------
        inputs: list of arrays
            the inputs to the model

        Returns
        -------
        (loss, outputs) where loss is the value of the model's loss function, and
        outputs is a list of the model's outputs
        """
        return self.learner.compute_model(inputs, self.learner.variables, False)

    def save_checkpoint(self,
                        max_checkpoints_to_keep: int = 5,
                        model_dir: Optional[str] = None) -> None:
        """Save a checkpoint to disk.

        Usually you do not need to call this method, since fit() saves checkpoints
        automatically.  If you have disabled automatic checkpointing during fitting,
        this can be called to manually write checkpoints.

        Parameters
        ----------
        max_checkpoints_to_keep: int
            the maximum number of checkpoints to keep.  Older checkpoints are discarded.
        model_dir: str, default None
            Model directory to save checkpoint to. If None, revert to self.model_dir
        """
        if model_dir is None:
            model_dir = self.model_dir
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)

        # Save the checkpoint to a file.

        data: Dict[str, Any] = {
            'model_state_dict':
                self.learner.__dict__,
            'learning_rate':
                self.learning_rate,
            'optimizer_state_dict':
                self._pytorch_optimizer.state_dict(),
            'task_optimizer_state_dict':
                self._pytorch_task_optimizer.state_dict(),
            'global_step':
                self._global_step
        }
        temp_file: str = os.path.join(model_dir, 'temp_checkpoint.pt')
        torch.save(data, temp_file)

        # Rename and delete older files.

        paths: List[str] = [
            os.path.join(model_dir, 'checkpoint%d.pt' % (i + 1))
            for i in range(max_checkpoints_to_keep)
        ]
        if os.path.exists(paths[-1]):
            os.remove(paths[-1])
        for i in reversed(range(max_checkpoints_to_keep - 1)):
            if os.path.exists(paths[i]):
                os.rename(paths[i], paths[i + 1])
        os.rename(temp_file, paths[0])

    def get_checkpoints(self, model_dir: Optional[str] = None) -> List[str]:
        """Get a list of all available checkpoint files.

        Parameters
        ----------
        model_dir: str, default None
            Directory to get list of checkpoints from. Reverts to self.model_dir if None

        """
        if model_dir is None:
            model_dir = self.model_dir
        files: List[str] = sorted(os.listdir(model_dir))
        files = [
            f for f in files if f.startswith('checkpoint') and f.endswith('.pt')
        ]
        return [os.path.join(model_dir, f) for f in files]



from os import path
from typing import Optional

import numpy as np
from rdkit.Chem import AllChem
from deepchem.utils import download_url, get_data_dir, untargz_file
from deepchem.utils.typing import RDKitMol
from deepchem.feat.base_classes import MolecularFeaturizer

DEFAULT_PRETRAINED_MODEL_URL = 'https://deepchemdata.s3-us-west-1.amazonaws.com/trained_models/mol2vec_model_300dim.tar.gz'


def _mol2alt_sentence(mol, radius):
    """Same as mol2sentence() except it only returns the alternating sentence
    Calc6ulates ECFP (Morgan fingerprint) and returns identifiers of substructures as 'sentence' (string).
    Returns a tuple with 1) a list with sentence for each radius and 2) a sentence with identifiers from all radii
    combined.
    NOTE: Words are ALWAYS reordered according to atom order in the input mol object.
    NOTE: Due to the way how Morgan FPs are generated, number of identifiers at each radius is smaller

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
    radius : float
        Fingerprint radius

    Returns
    -------
    list
        alternating sentence
    combined

    """
    # Copied from https://github.com/samoturk/mol2vec/blob/850d944d5f48a58e26ed0264332b5741f72555aa/mol2vec/features.py#L129-L168
    radii = list(range(int(radius) + 1))
    info = {}
    _ = AllChem.GetMorganFingerprint(
        mol, radius,
        bitInfo=info)  # info: dictionary identifier, atom_idx, radius

    mol_atoms = [a.GetIdx() for a in mol.GetAtoms()]
    dict_atoms = {x: {r: None for r in radii} for x in mol_atoms}

    for element in info:
        for atom_idx, radius_at in info[element]:
            dict_atoms[atom_idx][
                radius_at] = element  # {atom number: {fp radius: identifier}}

    # merge identifiers alternating radius to sentence: atom 0 radius0, atom 0 radius 1, etc.
    identifiers_alt = []
    for atom in dict_atoms:  # iterate over atoms
        for r in radii:  # iterate over radii
            identifiers_alt.append(dict_atoms[atom][r])

    alternating_sentence = map(str, [x for x in identifiers_alt if x])

    return list(alternating_sentence)


class Mol2VecFingerprint(MolecularFeaturizer):
    """Mol2Vec fingerprints.

    This class convert molecules to vector representations by using Mol2Vec.
    Mol2Vec is an unsupervised machine learning approach to learn vector representations
    of molecular substructures and the algorithm is based on Word2Vec, which is
    one of the most popular technique to learn word embeddings using neural network in NLP.
    Please see the details from [1]_.

    The Mol2Vec requires the pretrained model, so we use the model which is put on the mol2vec
    github repository [2]_. The default model was trained on 20 million compounds downloaded
    from ZINC using the following paramters.

    - radius 1
    - UNK to replace all identifiers that appear less than 4 times
    - skip-gram and window size of 10
    - embeddings size 300

    References
    ----------
    .. [1] Jaeger, Sabrina, Simone Fulle, and Samo Turk. "Mol2vec: unsupervised machine learning
        approach with chemical intuition." Journal of chemical information and modeling 58.1 (2018): 27-35.
    .. [2] https://github.com/samoturk/mol2vec/

    Note
    ----
    This class requires mol2vec to be installed.

    Examples
    --------
    >>> import deepchem as dc
    >>> from rdkit import Chem
    >>> smiles = ['CCC']
    >>> featurizer = dc.feat.Mol2VecFingerprint()
    >>> features = featurizer.featurize(smiles)
    >>> type(features)
    <class 'numpy.ndarray'>
    >>> features[0].shape
    (300,)

    """

    def __init__(self,
                 pretrain_model_path: Optional[str] = None,
                 radius: int = 1,
                 unseen: str = 'UNK'):
        """
        Parameters
        ----------
        pretrain_file: str, optional
            The path for pretrained model. If this value is None, we use the model which is put on
            github repository (https://github.com/samoturk/mol2vec/tree/master/examples/models).
            The model is trained on 20 million compounds downloaded from ZINC.
        radius: int, optional (default 1)
            The fingerprint radius. The default value was used to train the model which is put on
            github repository.
        unseen: str, optional (default 'UNK')
            The string to used to replace uncommon words/identifiers while training.

        """
        try:
            from gensim.models import word2vec
        except ModuleNotFoundError:
            raise ImportError("This class requires mol2vec to be installed.")

        self.radius = radius
        self.unseen = unseen
        self.mol2alt_sentence = _mol2alt_sentence
        if pretrain_model_path is None:
            data_dir = get_data_dir()
            pretrain_model_path = path.join(data_dir,
                                            'mol2vec_model_300dim.pkl')
            if not path.exists(pretrain_model_path):
                targz_file = path.join(data_dir, 'mol2vec_model_300dim.tar.gz')
                if not path.exists(targz_file):
                    download_url(DEFAULT_PRETRAINED_MODEL_URL, data_dir)
                untargz_file(path.join(data_dir, 'mol2vec_model_300dim.tar.gz'),
                             data_dir)
        # load pretrained models
        self.model = word2vec.Word2Vec.load(pretrain_model_path)

    def sentences2vec(self, sentences: list, model, unseen=None) -> np.ndarray:
        """Generate vectors for each sentence (list) in a list of sentences. Vector is simply a
        sum of vectors for individual words.

        Parameters
        ----------
        sentences : list, array
            List with sentences
        model : word2vec.Word2Vec
            Gensim word2vec model
        unseen : None, str
            Keyword for unseen words. If None, those words are skipped.
            https://stats.stackexchange.com/questions/163005/how-to-set-the-dictionary-for-text-analysis-using-neural-networks/163032#163032
        Returns
        -------
        np.array

        """
        keys = set(model.wv.key_to_index.keys())
        vec = []
        if unseen:
            unseen_vec = model.wv.get_vector(unseen)

        for sentence in sentences:
            if unseen:
                vec.append(
                    sum([
                        model.wv.get_vector(y) if y in set(sentence) &
                        keys else unseen_vec for y in sentence
                    ]))
            else:
                vec.append(
                    sum([
                        model.wv.get_vector(y)
                        for y in sentence
                        if y in set(sentence) & keys
                    ]))
        return np.array(vec)

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        """
        Calculate Mordred descriptors.

        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
            RDKit Mol object

        Returns
        -------
        np.ndarray
            1D array of mol2vec fingerprint. The default length is 300.

        """
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )
        sentence = self.mol2alt_sentence(datapoint, self.radius)
        feature = self.sentences2vec([sentence], self.model,
                                     unseen=self.unseen)[0]
        return feature



"""
Implementation of the Ferminet class in pytorch
"""

from typing import List, Optional, Tuple
# import torch.nn as nn
from rdkit import Chem
import numpy as np
from deepchem.utils.molecule_feature_utils import ALLEN_ELECTRONEGATIVTY
from deepchem.models.torch_models import TorchModel
import torch
from deepchem.models.torch_models.layers import FerminetElectronFeature, FerminetEnvelope

from deepchem.utils.electron_sampler import ElectronSampler


class Ferminet(torch.nn.Module):
    """A deep-learning based Variational Monte Carlo method [1]_ for calculating the ab-initio
    solution of a many-electron system.

    This model must be pre-trained on the HF baseline and then can be used to improve on it using electronic interactions
    via the learned electron distance features. Ferminet models the orbital values using envelope functions and the
    calculated electron features, which can then be used to approximate the wavefunction for better sampling of electrons.

    Example
    -------
    >>> import numpy as np
    >>> import deepchem as dc
    >>> import torch
    >>> H2_molecule =  torch.Tensor([[0, 0, 0.748], [0, 0, 0]])
    >>> H2_charge = torch.Tensor([[1], [1]])
    >>> model = dc.models.Ferminet(nucleon_pos=H2_molecule, nuclear_charge=H2_charge, batch_size=1)
    >>> electron = np.random.rand(1, 2*3)
    >>> wavefunction = model.forward(electron)

    References
    ----------
    .. [1] Spencer, James S., et al. Better, Faster Fermionic Neural Networks. arXiv:2011.07125, arXiv, 13 Nov. 2020. arXiv.org, http://arxiv.org/abs/2011.07125.

    """

    def __init__(self,
                 nucleon_pos: torch.Tensor,
                 nuclear_charge: torch.Tensor,
                 spin: tuple,
                 n_one: List = [256, 256, 256, 256],
                 n_two: List = [32, 32, 32, 32],
                 determinant: int = 16,
                 batch_size: int = 8) -> None:
        """
        Parameters:
        -----------
        nucleon_pos: torch.Tensor
            tensor containing the nucleons coordinates, it is in the shape of (number of atoms, 3)
        nuclear_charge: torch.Tensor
            tensor containing the electron number associated with each nucleon, it is in the shape of (number of atoms, no_of_electron)
        n_one: List
            List of hidden units for the one-electron stream in each layer
        n_two: List
            List of hidden units for the two-electron stream in each layer
        determinant: int
            Number of determinants for the final solution
        batch_size: int
            Number of molecule samples to be in each batch

        Attributes
        ----------
        running_diff: torch.Tensor
            torch tensor containing the loss which gets updated for each random walk performed
        ferminet_layer: torch.nn.ModuleList
            Modulelist containing the ferminet electron feature layer
        ferminet_layer_envelope: torch.nn.ModuleList
            Modulelist containing the ferminet envelope electron feature layer
        nuclear_nuclear_potential: torch.Tensor
            Torch tensor containing the inter-nuclear potential in the molecular system
        """
        super(Ferminet, self).__init__()
        if len(n_one) != len(n_two):
            raise ValueError(
                "The number of layers in one-electron and two-electron stream should be equal"
            )
        else:
            self.layers = len(n_one)

        self.nucleon_pos = nucleon_pos
        self.determinant = determinant
        self.batch_size = batch_size
        self.spin = spin
        self.total_electron = spin[0] + spin[1]
        self.nuclear_charge = nuclear_charge
        self.n_one = n_one
        self.n_two = n_two
        self.ferminet_layer: torch.nn.ModuleList = torch.nn.ModuleList()
        self.ferminet_layer_envelope: torch.nn.ModuleList = torch.nn.ModuleList(
        )
        self.running_diff: torch.Tensor = torch.zeros(self.batch_size)
        self.nuclear_nuclear_potential: torch.Tensor = self.calculate_nuclear_nuclear(
        )

        self.ferminet_layer.append(
            FerminetElectronFeature(self.n_one, self.n_two,
                                    self.nucleon_pos.size()[0], self.batch_size,
                                    self.total_electron,
                                    [self.spin[0], self.spin[1]]))
        self.ferminet_layer_envelope.append(
            FerminetEnvelope(self.n_one, self.n_two, self.total_electron,
                             self.batch_size, [self.spin[0], self.spin[1]],
                             self.nucleon_pos.size()[0], self.determinant))

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """
        forward function

        Parameters:
        -----------
        input: torch.Tensor
            contains the sampled electrons coordinate in the shape (batch_size,number_of_electrons*3)

        Returns:
        --------
        psi: torch.Tensor
            contains the wavefunction - 'psi' value. It is in the shape (batch_size), where each row corresponds to the solution of one of the batches
        """
        # creating one and two electron features
        eps = torch.tensor(1e-36)
        self.input = input.reshape((self.batch_size, -1, 3))
        two_electron_vector = self.input.unsqueeze(1) - self.input.unsqueeze(2)
        two_electron_distance = torch.linalg.norm(two_electron_vector + eps,
                                                  dim=3).unsqueeze(3)
        two_electron = torch.cat((two_electron_vector, two_electron_distance),
                                 dim=3)
        two_electron = torch.reshape(
            two_electron,
            (self.batch_size, self.total_electron, self.total_electron, -1))
        one_electron_vector = self.input.unsqueeze(
            1) - self.nucleon_pos.unsqueeze(1)
        one_electron_distance = torch.linalg.norm(one_electron_vector, dim=3)
        one_electron = torch.cat(
            (one_electron_vector, one_electron_distance.unsqueeze(-1)), dim=3)
        one_electron = torch.reshape(one_electron.permute(0, 2, 1, 3),
                                     (self.batch_size, self.total_electron, -1))
        one_electron_vector_permuted = one_electron_vector.permute(0, 2, 1,
                                                                   3).float()

        one_electron, _ = self.ferminet_layer[0].forward(
            one_electron.to(torch.float32), two_electron.to(torch.float32))
        self.psi, self.psi_up, self.psi_down = self.ferminet_layer_envelope[
            0].forward(one_electron, one_electron_vector_permuted)
        return self.psi

    def loss(self,
             psi_up_mo: List[Optional[np.ndarray]] = [None],
             psi_down_mo: List[Optional[np.ndarray]] = [None],
             pretrain: List[bool] = [True]):
        """
        Implements the loss function for both pretraining and the actual training parts.

        Parameters
        ----------
        psi_up_mo: List[Optional[np.ndarray]] (default [None])
            numpy array containing the sampled hartreee fock up-spin orbitals
        psi_down_mo: List[Optional[np.ndarray]] (default [None])
            numpy array containing the sampled hartreee fock down-spin orbitals
        pretrain: List[bool] (default [True])
            indicates whether the model is pretraining
        """
        criterion = torch.nn.MSELoss()
        if pretrain[0]:
            psi_up_mo_torch = torch.from_numpy(psi_up_mo).unsqueeze(1)
            psi_down_mo_torch = torch.from_numpy(psi_down_mo).unsqueeze(1)
            self.running_diff = self.running_diff.float() + criterion(
                self.psi_up.float(),
                psi_up_mo_torch.float()).float() + criterion(
                    self.psi_down.float(), psi_down_mo_torch.float()).float()
        else:
            energy = self.calculate_electron_electron(
            ) - self.calculate_electron_nuclear(
            ) + self.nuclear_nuclear_potential + self.calculate_kinetic_energy(
            )
            return energy.detach()

    def calculate_nuclear_nuclear(self,) -> torch.Tensor:
        """
        Function to calculate where only the nucleus terms are involved and does not change when new electrons are sampled.
        atom-atom potential term = Zi*Zj/|Ri-Rj|, where Zi, Zj are the nuclear charges and Ri, Rj are nuclear coordinates

        Returns:
        --------
        A torch tensor of a scalar value containing the nuclear-nuclear potential term (does not change for the molecule system with sampling of electrons)
        """

        potential = torch.nan_to_num(
            (self.nuclear_charge * 1 /
             torch.cdist(self.nucleon_pos.float(), self.nucleon_pos.float()) *
             self.nuclear_charge.unsqueeze(1)),
            posinf=0.0,
            neginf=0.0)
        potential = (torch.sum(potential) / 2).unsqueeze(0)
        return potential.detach()

    def calculate_electron_nuclear(self,):
        """
        Function to calculate the expected electron-nuclear potential term per batch
        nuclear-electron potential term = Zi/|Ri-rj|, rj is the electron coordinates, Ri is the nuclear coordinates, Zi is the nuclear charge

        Returns:
        --------
        A torch tensor of a scalar value containing the electron-nuclear potential term.
        """

        potential = torch.sum(torch.sum(
            (1 / torch.cdist(self.input.float(), self.nucleon_pos.float())) *
            self.nuclear_charge,
            axis=-1),
                              axis=-1)
        return potential.detach()

    def calculate_electron_electron(self,):
        """
        Function to calculate the expected electron-nuclear potential term per batch
        nuclear-electron potential term = 1/|ri-rj|, ri, rj is the electron coordinates

        Returns:
        --------
        A torch tensor of a scalar value containing the electron-electron potential term.
        """
        potential = torch.sum(torch.sum(torch.nan_to_num(
            (1 / torch.cdist(self.input.float(), self.input.float())),
            posinf=0.0,
            neginf=0.0),
                                        axis=-1),
                              axis=-1) / 2
        return potential.detach()

    def calculate_kinetic_energy(self,):
        """
        Function to calculate the expected kinetic energy term per batch
        It is calculated via:
        \sum_{ri}^{}[(\pdv[]{log|\Psi|}{(ri)})^2 + \pdv[2]{log|\Psi|}{(ri)}]

        Returns:
        --------
        A torch tensor of a scalar value containing the electron-electron potential term.
        """
        # using functorch to calcualte hessian and jacobian in one go
        # using index tensors to index out the hessian elemennts corresponding to the same variable (cross-variable derivatives are ignored)
        i = torch.arange(self.batch_size).view(self.batch_size, 1, 1, 1, 1, 1,
                                               1)
        j = torch.arange(self.total_electron).view(1, self.total_electron, 1, 1,
                                                   1, 1, 1)
        k = torch.arange(3).view(1, 1, 3, 1, 1, 1, 1)

        # doing all the calculation and detaching from graph to save memory, which allows larger batch size
        jacobian_square_sum = torch.sum(torch.sum(torch.sum(torch.pow(
            torch.func.jacrev(lambda x: torch.log(torch.abs(self.forward(x))))(
                self.input), 2),
                                                            axis=-1),
                                                  axis=-1),
                                        axis=-1).detach()
        hessian_sum = torch.sum(torch.reshape(
            torch.func.hessian(lambda x: torch.log(torch.abs(self.forward(x))))(
                self.input)[i, i, j, k, i, j, k],
            (self.batch_size, self.total_electron, 3)).detach(),
                                axis=(1, 2))
        kinetic_energy = -1 * 0.5 * (jacobian_square_sum + hessian_sum)
        return kinetic_energy


class FerminetModel(TorchModel):
    """A deep-learning based Variational Monte Carlo method [1]_ for calculating the ab-initio
    solution of a many-electron system.

    This model aims to calculate the ground state energy of a multi-electron system
    using a baseline solution as the Hartree-Fock. An MCMC technique is used to sample
    electrons and DNNs are used to caluclate the square magnitude of the wavefunction,
    in which electron-electron repulsions also are included in the calculation(in the
    form of Jastrow factor envelopes). The model requires only the nucleus' coordinates
    as input.

    This method is based on the following paper:

    Example
    -------
    >>> from deepchem.models.torch_models.Ferminet import FerminetModel
    >>> H2_molecule = [['H', [0, 0, 0]], ['H', [0, 0, 0.748]]]
    >>> mol = FerminetModel(H2_molecule, spin=0, ion_charge=0, training='pretraining')
    >>> mol.train(nb_epoch=3)
    >>> print(mol.model.psi_up.size())
    torch.Size([1, 1])

    References
    ----------
    .. [1] Spencer, James S., et al. Better, Faster Fermionic Neural Networks. arXiv:2011.07125, arXiv, 13 Nov. 2020. arXiv.org, http://arxiv.org/abs/2011.07125.

    Note
    ----
    This class requires pySCF to be installed.
    """

    def __init__(self,
                 nucleon_coordinates: List[List],
                 spin: int,
                 ion_charge: int,
                 seed: Optional[int] = None,
                 batch_no: int = 8,
                 random_walk_steps: int = 10,
                 steps_per_update: int = 10,
                 tasks: str = 'pretraining'):
        """
    Parameters:
    -----------
    nucleon_coordinates: List[List]
      A list containing nucleon coordinates as the values with the keys as the element's symbol.
    spin: int
      The total spin of the molecule system.
    ion_charge: int
      The total charge of the molecule system.
    seed_no: int, optional (default None)
      Random seed to use for electron initialization.
    batch_no: int, optional (default 8)
      Number of batches of the electron's positions to be initialized.
    random_walk_steps: int (default 10)
        Number of random walk steps to be performed in a single move.
    steps_per_update: int (default: 10)
        Number of steps after which the electron sampler should update the electron parameters.
    tasks: str (default: 'pretraining')
        The type of task to be performed - 'pretraining', 'training'

    Attributes:
    -----------
    nucleon_pos: np.ndarray
        numpy array value of nucleon_coordinates
    electron_no: np.ndarray
        Torch tensor containing electrons for each atom in the nucleus
    molecule: ElectronSampler
        ElectronSampler object which performs MCMC and samples electrons
    loss_value: torch.Tensor (default torch.tensor(0))
        torch tensor storing the loss value from the last iteration
    """
        self.nucleon_coordinates = nucleon_coordinates
        self.seed = seed
        self.batch_no = batch_no
        self.spin = spin
        self.ion_charge = ion_charge
        self.batch_no = batch_no
        self.random_walk_steps = random_walk_steps
        self.steps_per_update = steps_per_update
        self.loss_value: torch.Tensor = torch.tensor(0)
        self.tasks = tasks

        no_electrons = []
        nucleons = []
        electronegativity = []

        table = Chem.GetPeriodicTable()
        index = 0
        for i in self.nucleon_coordinates:
            atomic_num = table.GetAtomicNumber(i[0])
            electronegativity.append([index, ALLEN_ELECTRONEGATIVTY[i[0]]])
            no_electrons.append([atomic_num])
            nucleons.append(i[1])
            index += 1

        self.electron_no: np.ndarray = np.array(no_electrons)
        charge: np.ndarray = self.electron_no.reshape(
            np.shape(self.electron_no)[0])
        self.nucleon_pos: np.ndarray = np.array(nucleons)
        electro_neg = np.array(electronegativity)

        # Initialization for ionic molecules
        if np.sum(self.electron_no) < self.ion_charge:
            raise ValueError("Given charge is not initializable")

        # Initialization for ionic molecules
        if self.ion_charge != 0:
            if len(nucleons
                  ) == 1:  # for an atom, directly the charge is applied
                self.electron_no[0][0] -= self.ion_charge
            else:  # for a multiatomic molecule, the most electronegative atom gets a charge of -1 and vice versa. The remaining charges are assigned in terms of decreasing(for anionic charge) and increasing(for cationic charge) electronegativity.
                electro_neg = electro_neg[electro_neg[:, 1].argsort()]
                if self.ion_charge > 0:
                    for iter in range(self.ion_charge):
                        self.electron_no[int(electro_neg[iter][0])][0] -= 1
                else:
                    for iter in range(-self.ion_charge):
                        self.electron_no[int(electro_neg[-1 - iter][0])][0] += 1

        total_electrons = np.sum(self.electron_no)

        if self.spin >= 0:
            self.up_spin = (total_electrons + 2 * self.spin) // 2
            self.down_spin = total_electrons - self.up_spin
        else:
            self.down_spin = (total_electrons - 2 * self.spin) // 2
            self.up_spin = total_electrons - self.down_spin

        if self.up_spin - self.down_spin != self.spin:
            raise ValueError("Given spin is not feasible")

        nucl = torch.from_numpy(self.nucleon_pos)
        self.model = Ferminet(nucl,
                              spin=(self.up_spin, self.down_spin),
                              nuclear_charge=torch.Tensor(charge),
                              batch_size=self.batch_no)

        self.molecule: ElectronSampler = ElectronSampler(
            batch_no=self.batch_no,
            central_value=self.nucleon_pos,
            seed=self.seed,
            f=lambda x: self.random_walk(x
                                        ),  # Will be replaced in successive PR
            steps=self.random_walk_steps,
            steps_per_update=self.steps_per_update
        )  # sample the electrons using the electron sampler
        self.molecule.gauss_initialize_position(
            self.electron_no,
            stddev=1.0)  # initialize the position of the electrons
        self.prepare_hf_solution()
        super(FerminetModel, self).__init__(
            self.model,
            loss=torch.nn.MSELoss())  # will update the loss in successive PR

    def evaluate_hf(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Helper function to calculate orbital values at sampled electron's position.

        Parameters:
        -----------
        x: np.ndarray
            Contains the sampled electrons coordinates in a numpy array.

        Returns:
        --------
        2 numpy arrays containing the up-spin and down-spin orbitals in a numpy array respectively.
        """
        x = np.reshape(x, [-1, 3 * (self.up_spin + self.down_spin)])
        leading_dims = x.shape[:-1]
        x = np.reshape(x, [-1, 3])
        coeffs = self.mf.mo_coeff
        gto_op = 'GTOval_sph'
        ao_values = self.mol.eval_gto(gto_op, x)
        mo_values = tuple(np.matmul(ao_values, coeff) for coeff in coeffs)
        mo_values_list = [
            np.reshape(mo, leading_dims + (self.up_spin + self.down_spin, -1))
            for mo in mo_values
        ]
        return mo_values_list[0][
            ..., :self.up_spin, :self.up_spin], mo_values_list[1][
                ..., self.up_spin:, :self.down_spin]

    def prepare_hf_solution(self):
        """Prepares the HF solution for the molecule system which is to be used in pretraining
        """
        try:
            import pyscf
        except ModuleNotFoundError:
            raise ImportError("This module requires pySCF")

        molecule = ""
        for i in range(len(self.nucleon_pos)):
            molecule = molecule + self.nucleon_coordinates[i][0] + " " + str(
                self.nucleon_coordinates[i][1][0]) + " " + str(
                    self.nucleon_coordinates[i][1][1]) + " " + str(
                        self.nucleon_coordinates[i][1][2]) + ";"
        self.mol = pyscf.gto.Mole(atom=molecule, basis='sto-3g')
        self.mol.parse_arg = False
        self.mol.unit = 'Bohr'
        self.mol.spin = (self.up_spin - self.down_spin)
        self.mol.charge = self.ion_charge
        self.mol.build(parse_arg=False)
        self.mf = pyscf.scf.UHF(self.mol)
        _ = self.mf.kernel()

    def random_walk(self, x: np.ndarray):
        """
        Function to be passed on to electron sampler for random walk and gets called at each step of sampling

        Parameters
        ----------
        x: np.ndarray
            contains the sampled electrons coordinate in the shape (batch_size,number_of_electrons*3)

        Returns:
        --------
        A numpy array containing the joint probability of the hartree fock and the sampled electron's position coordinates
        """
        x_torch = torch.from_numpy(x).view(self.batch_no, -1, 3)
        if self.tasks == 'pretraining':
            x_torch.requires_grad = True
        else:
            x_torch.requires_grad = False
        output = self.model.forward(x_torch)
        np_output = output.detach().cpu().numpy()

        if self.tasks == 'pretraining':
            up_spin_mo, down_spin_mo = self.evaluate_hf(x)
            hf_product = np.prod(
                np.diagonal(up_spin_mo, axis1=1, axis2=2), axis=1) * np.prod(
                    np.diagonal(down_spin_mo, axis1=1, axis2=2), axis=1)
            self.model.loss(up_spin_mo, down_spin_mo)
            np_output[:int(self.batch_no / 2)] = hf_product[:int(self.batch_no /
                                                                 2)]
            return 2 * np.log(np.abs(np_output))

        if self.tasks == 'burn':
            return 2 * np.log(np.abs(np_output))

        if self.tasks == 'training':
            energy = self.model.loss(pretrain=[False])
            self.energy_sampled: torch.Tensor = torch.cat(
                (self.energy_sampled, energy.unsqueeze(0)))
            return 2 * np.log(np.abs(np_output))

    def prepare_train(self, burn_in: int = 100):
        """
        Function to perform burn-in and to change the model parameters for training.

        Parameters
        ----------
        burn_in:int (default: 100)
            number of steps for to perform burn-in before the aactual training.
        """
        self.tasks = 'burn'
        self.molecule.gauss_initialize_position(self.electron_no, stddev=1.0)
        tmp_x = self.molecule.x
        for _ in range(burn_in):
            self.molecule.move(stddev=0.02)
            self.molecule.x = tmp_x
        self.molecule.move(stddev=0.02)
        self.tasks = 'training'

    def train(self,
              nb_epoch: int = 200,
              lr: float = 0.002,
              weight_decay: float = 0,
              std: float = 0.08,
              std_init: float = 0.02,
              steps_std: int = 100):
        """
        function to run training or pretraining.

        Parameters
        ----------
        nb_epoch: int (default: 200)
            contains the number of pretraining steps to be performed
        lr : float (default: 0.002)
            contains the learning rate for the model fitting
        weight_decay: float (default: 0.002)
            contains the weight_decay for the model fitting
        std: float (default: 0.08)
            The standard deviation for the electron update during training
        std_init: float (default: 0.02)
            The standard deviation for the electron update during pretraining
        steps_std: float (default 100)
            The number of steps for standard deviation increase
        """

        # hook function below is an efficient way modifying the gradients on the go rather than looping
        def energy_hook(grad, random_walk_steps):
            """
            hook function to modify the gradients
            """
            # using non-local variables as a means of parameter passing
            nonlocal energy_local, energy_mean
            new_grad = (2 / random_walk_steps) * (
                (energy_local - energy_mean) * grad)
            return new_grad.float()

        optimizer = torch.optim.Adam(self.model.parameters(),
                                     lr=lr,
                                     weight_decay=weight_decay)

        if (self.tasks == 'pretraining'):
            for iteration in range(nb_epoch):
                optimizer.zero_grad()
                accept = self.molecule.move(stddev=std_init)
                if iteration % steps_std == 0:
                    if accept > 0.55:
                        std_init *= 1.1
                    else:
                        std_init /= 1.1
                self.loss_value = (torch.mean(self.model.running_diff) /
                                   self.random_walk_steps)
                self.loss_value.backward()
                optimizer.step()
                self.model.running_diff = torch.zeros(self.batch_no)

        if (self.tasks == 'training'):
            energy_local = None
            optimizer = torch.optim.Adam(self.model.parameters(),
                                         lr=lr,
                                         weight_decay=weight_decay)
            self.final_energy = torch.tensor(0.0)
            with torch.no_grad():
                hooks = list(
                    map(
                        lambda param: param.
                        register_hook(lambda grad: energy_hook(
                            grad, self.random_walk_steps)),
                        self.model.parameters()))
            for iteration in range(nb_epoch):
                optimizer.zero_grad()
                self.energy_sampled = torch.tensor([])
                # the move function calculates the energy of sampled electrons and samples new set of electrons (does not calculate loss)
                accept = self.molecule.move(stddev=std)
                if iteration % steps_std == 0:
                    if accept > 0.55:
                        std_init *= 1.2
                    else:
                        std_init /= 1.2
                median, _ = torch.median(self.energy_sampled, 0)
                variance = torch.mean(torch.abs(self.energy_sampled - median))
                # clipping local energies which are away 5 times the variance from the median
                clamped_energy = torch.clamp(self.energy_sampled,
                                             max=median + 5 * variance,
                                             min=median - 5 * variance)
                energy_mean = torch.mean(clamped_energy)
                self.final_energy = self.final_energy + energy_mean
                # using the sampled electrons from the electron sampler for bacward pass and modifying gradients
                sample_history = torch.from_numpy(
                    self.molecule.sampled_electrons).view(
                        self.random_walk_steps, self.batch_no, -1, 3)
                optimizer.zero_grad()
                for i in range(self.random_walk_steps):
                    # going through each step of random walk and calculating the modified gradients with local energies
                    input_electron = sample_history[i]
                    input_electron.requires_grad = True
                    energy_local = torch.mean(clamped_energy[i])
                    self.model.forward(input_electron)
                    self.loss_value = torch.mean(
                        torch.log(torch.abs(self.model.psi)))
                    self.loss_value.backward()
                optimizer.step()
            self.final_energy = self.final_energy / nb_epoch
            list(map(lambda hook: hook.remove(), hooks))



"""
Miscellaneous utility functions.
"""
# flake8: noqa
import logging

logger = logging.getLogger(__name__)

from deepchem.utils.misc_utils import indent
from deepchem.utils.misc_utils import shape2str

from deepchem.utils.batch_utils import batch_coulomb_matrix_features

from deepchem.utils.attribute_utils import set_attr
from deepchem.utils.attribute_utils import get_attr
from deepchem.utils.attribute_utils import del_attr

from deepchem.utils.conformers import ConformerGenerator
from deepchem.utils.evaluate import relative_difference
from deepchem.utils.evaluate import Evaluator
from deepchem.utils.evaluate import GeneratorEvaluator

from deepchem.utils.coordinate_box_utils import CoordinateBox
from deepchem.utils.coordinate_box_utils import intersect_interval
from deepchem.utils.coordinate_box_utils import intersection
from deepchem.utils.coordinate_box_utils import union
from deepchem.utils.coordinate_box_utils import merge_overlapping_boxes
from deepchem.utils.coordinate_box_utils import get_face_boxes

from deepchem.utils.data_utils import pad_array
from deepchem.utils.data_utils import get_data_dir
from deepchem.utils.data_utils import download_url
from deepchem.utils.data_utils import untargz_file
from deepchem.utils.data_utils import unzip_file
from deepchem.utils.data_utils import UniversalNamedTemporaryFile
from deepchem.utils.data_utils import load_image_files
from deepchem.utils.data_utils import load_sdf_files
from deepchem.utils.data_utils import load_csv_files
from deepchem.utils.data_utils import load_json_files
from deepchem.utils.data_utils import load_pickle_files
from deepchem.utils.data_utils import load_data
from deepchem.utils.data_utils import save_to_disk
from deepchem.utils.data_utils import load_from_disk
from deepchem.utils.data_utils import save_dataset_to_disk
from deepchem.utils.data_utils import load_dataset_from_disk
from deepchem.utils.data_utils import remove_missing_entries

from deepchem.utils.debug_utils import get_print_threshold
from deepchem.utils.debug_utils import set_print_threshold
from deepchem.utils.debug_utils import get_max_print_size
from deepchem.utils.debug_utils import set_max_print_size

from deepchem.utils.fragment_utils import AtomShim
from deepchem.utils.fragment_utils import MolecularFragment
from deepchem.utils.fragment_utils import get_partial_charge
from deepchem.utils.fragment_utils import merge_molecular_fragments
from deepchem.utils.fragment_utils import get_mol_subset
from deepchem.utils.fragment_utils import strip_hydrogens
from deepchem.utils.fragment_utils import get_contact_atom_indices
from deepchem.utils.fragment_utils import reduce_molecular_complex_to_contacts

from deepchem.utils.genomics_utils import seq_one_hot_encode
from deepchem.utils.genomics_utils import encode_bio_sequence

from deepchem.utils.geometry_utils import unit_vector
from deepchem.utils.geometry_utils import angle_between
from deepchem.utils.geometry_utils import generate_random_unit_vector
from deepchem.utils.geometry_utils import generate_random_rotation_matrix
from deepchem.utils.geometry_utils import is_angle_within_cutoff
from deepchem.utils.geometry_utils import compute_centroid
from deepchem.utils.geometry_utils import subtract_centroid
from deepchem.utils.geometry_utils import compute_protein_range
from deepchem.utils.geometry_utils import compute_pairwise_distances

from deepchem.utils.graph_utils import fourier_encode_dist
from deepchem.utils.graph_utils import aggregate_mean
from deepchem.utils.graph_utils import aggregate_max
from deepchem.utils.graph_utils import aggregate_min
from deepchem.utils.graph_utils import aggregate_std
from deepchem.utils.graph_utils import aggregate_var
from deepchem.utils.graph_utils import aggregate_moment
from deepchem.utils.graph_utils import aggregate_sum
from deepchem.utils.graph_utils import scale_identity
from deepchem.utils.graph_utils import scale_amplification
from deepchem.utils.graph_utils import scale_attenuation

from deepchem.utils.hash_utils import hash_ecfp
from deepchem.utils.hash_utils import hash_ecfp_pair
from deepchem.utils.hash_utils import vectorize

from deepchem.utils.molecule_feature_utils import one_hot_encode
from deepchem.utils.molecule_feature_utils import get_atom_type_one_hot
from deepchem.utils.molecule_feature_utils import construct_hydrogen_bonding_info
from deepchem.utils.molecule_feature_utils import get_atom_hydrogen_bonding_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_is_in_aromatic_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_hybridization_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_total_num_Hs_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_chirality_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_formal_charge
from deepchem.utils.molecule_feature_utils import get_atom_formal_charge_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_partial_charge
from deepchem.utils.molecule_feature_utils import get_atom_total_degree_one_hot
from deepchem.utils.molecule_feature_utils import get_bond_type_one_hot
from deepchem.utils.molecule_feature_utils import get_bond_is_in_same_ring_one_hot
from deepchem.utils.molecule_feature_utils import get_bond_is_conjugated_one_hot
from deepchem.utils.molecule_feature_utils import get_bond_stereo_one_hot
from deepchem.utils.molecule_feature_utils import get_bond_graph_distance_one_hot

from deepchem.utils.pdbqt_utils import pdbqt_to_pdb
from deepchem.utils.pdbqt_utils import convert_protein_to_pdbqt
from deepchem.utils.pdbqt_utils import convert_mol_to_pdbqt

from deepchem.utils.docking_utils import write_vina_conf
from deepchem.utils.docking_utils import write_gnina_conf
from deepchem.utils.docking_utils import read_gnina_log
from deepchem.utils.docking_utils import load_docked_ligands
from deepchem.utils.docking_utils import prepare_inputs

from deepchem.utils.voxel_utils import convert_atom_to_voxel
from deepchem.utils.voxel_utils import convert_atom_pair_to_voxel
from deepchem.utils.voxel_utils import voxelize

from deepchem.utils.sequence_utils import hhblits
from deepchem.utils.sequence_utils import hhsearch

try:
    from deepchem.utils.periodic_table_utils import get_atomz
    from deepchem.utils.periodic_table_utils import get_atom_mass
    from deepchem.utils.periodic_table_utils import get_period

    from deepchem.utils.pytorch_utils import unsorted_segment_sum
    from deepchem.utils.pytorch_utils import segment_sum
    from deepchem.utils.pytorch_utils import chunkify
    from deepchem.utils.pytorch_utils import get_memory
    from deepchem.utils.pytorch_utils import gaussian_integral

    from deepchem.utils.safeops_utils import safepow
    from deepchem.utils.safeops_utils import safenorm
    from deepchem.utils.safeops_utils import occnumber
    from deepchem.utils.safeops_utils import get_floor_and_ceil
    from deepchem.utils.safeops_utils import safe_cdist
except ModuleNotFoundError as e:
    logger.warning(
        f'Skipped loading some Pytorch utilities, missing a dependency. {e}')



import sys
import time
from setuptools import setup, find_packages

if '--release' in sys.argv:
    IS_RELEASE = True
    sys.argv.remove('--release')
else:
    # Build a nightly package by default.
    IS_RELEASE = False

# Environment-specific dependencies.
extras = {
    'jax': ['jax', 'jaxlib', 'dm-haiku', 'optax'],
    'torch': ['torch', 'torchvision', 'pytorch-lightning', 'dgl', 'dgllife'],
    'tensorflow': ['tensorflow', 'tensorflow_probability', 'tensorflow_addons'],
    'dqc': ['dqc', 'xitorch', 'torch', 'pylibxc2']
}


# get the version from deepchem/__init__.py
def _get_version():
    with open('deepchem/__init__.py') as fp:
        for line in fp:
            if line.startswith('__version__'):
                g = {}
                exec(line, g)
                base = g['__version__']
                if IS_RELEASE:
                    return base
                else:
                    # nightly version : .devYearMonthDayHourMinute
                    if base.endswith('.dev') is False:
                        # Force to add `.dev` if `--release` option isn't passed when building
                        base += '.dev'
                    return base + time.strftime("%Y%m%d%H%M%S")

        raise ValueError('`__version__` not defined in `deepchem/__init__.py`')


setup(name='deepchem',
      version=_get_version(),
      url='https://github.com/deepchem/deepchem',
      maintainer='DeepChem contributors',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: Information Technology',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
      ],
      license='MIT',
      description='Deep learning models for drug discovery, \
        quantum chemistry, and the life sciences.',
      keywords=[
          'deepchem',
          'chemistry',
          'biology',
          'materials-science',
          'life-science',
          'drug-discovery',
      ],
      packages=find_packages(exclude=["*.tests"]),
      project_urls={
          'Documentation': 'https://deepchem.readthedocs.io/en/latest/',
          'Source': 'https://github.com/deepchem/deepchem',
      },
      install_requires=[
          'joblib',
          'numpy>=1.21',
          'pandas',
          'scikit-learn',
          'sympy',
          'scipy>=1.10.1',
          'rdkit',
      ],
      extras_require=extras,
      python_requires='>=3.7,<3.11')



"""Utility functions for working with PyTorch."""

import scipy
import torch
from typing import Callable, Union, List, Generator, Tuple


def get_activation(fn: Union[Callable, str]):
    """Get a PyTorch activation function, specified either directly or as a string.

    This function simplifies allowing users to specify activation functions by name.
    If a function is provided, it is simply returned unchanged.  If a string is provided,
    the corresponding function in torch.nn.functional is returned.
    """
    if isinstance(fn, str):
        return getattr(torch.nn.functional, fn)
    return fn


def unsorted_segment_sum(data: torch.Tensor, segment_ids: torch.Tensor,
                         num_segments: int) -> torch.Tensor:
    """Computes the sum along segments of a tensor. Analogous to tf.unsorted_segment_sum.

    Parameters
    ----------
    data: torch.Tensor
        A tensor whose segments are to be summed.
    segment_ids: torch.Tensor
        The segment indices tensor.
    num_segments: int
        The number of segments.

    Returns
    -------
    tensor: torch.Tensor

    Examples
    --------
    >>> segment_ids = torch.Tensor([0, 1, 0]).to(torch.int64)
    >>> data = torch.Tensor([[1, 2, 3, 4], [5, 6, 7, 8], [4, 3, 2, 1]])
    >>> num_segments = 2
    >>> result = unsorted_segment_sum(data=data,
    ...                               segment_ids=segment_ids,
    ...                               num_segments=num_segments)
    >>> data.shape[0]
    3
    >>> segment_ids.shape[0]
    3
    >>> len(segment_ids.shape)
    1
    >>> result
    tensor([[5., 5., 5., 5.],
            [5., 6., 7., 8.]])

    """

    if len(segment_ids.shape) != 1:
        raise AssertionError("segment_ids have be a 1-D tensor")

    if data.shape[0] != segment_ids.shape[0]:
        raise AssertionError(
            "segment_ids should be the same size as dimension 0 of input.")

    s = torch.prod(torch.tensor(data.shape[1:])).long()
    segment_ids = segment_ids.repeat_interleave(s).view(segment_ids.shape[0],
                                                        *data.shape[1:])

    # data.shape and segment_ids.shape should be equal
    assert data.shape == segment_ids.shape
    shape: List[int] = [num_segments] + list(data.shape[1:])
    tensor: torch.Tensor = torch.zeros(*shape).scatter_add(
        0, segment_ids, data.float())
    tensor = tensor.type(data.dtype)
    return tensor


def segment_sum(data: torch.Tensor, segment_ids: torch.Tensor) -> torch.Tensor:
    """ This function computes the sum of values along segments within a tensor. It is useful when you have a tensor with segment IDs and you want to compute the sum of values for each segment.
    This function is analogous to tf.segment_sum. (https://www.tensorflow.org/api_docs/python/tf/math/segment_sum).

    Parameters
    ----------
    data: torch.Tensor
        A pytorch tensor containing the values to be summed. It can have any shape, but its rank (number of dimensions) should be at least 1.
    segment_ids: torch.Tensor
        A 1-D tensor containing the indices for the segmentation. The segments can be any non-negative integer values, but they must be sorted in non-decreasing order.

    Returns
    -------
    out_tensor: torch.Tensor
        Tensor with the same shape as data, where each value corresponds to the sum of values within the corresponding segment.

    Examples
    --------
    >>> data = torch.Tensor([[1, 2, 3, 4], [4, 3, 2, 1], [5, 6, 7, 8]])
    >>> segment_ids = torch.Tensor([0, 0, 1]).to(torch.int64)
    >>> result = segment_sum(data=data, segment_ids=segment_ids)
    >>> data.shape[0]
    3
    >>> segment_ids.shape[0]
    3
    >>> len(segment_ids.shape)
    1
    >>> result
    tensor([[5., 5., 5., 5.],
            [5., 6., 7., 8.]])

    """
    if not all(segment_ids[i] <= segment_ids[i + 1]
               for i in range(len(segment_ids) - 1)):
        raise AssertionError("elements of segment_ids must be sorted")

    if len(segment_ids.shape) != 1:
        raise AssertionError("segment_ids have be a 1-D tensor")

    if data.shape[0] != segment_ids.shape[0]:
        raise AssertionError(
            "segment_ids should be the same size as dimension 0 of input.")

    num_segments = len(torch.unique(segment_ids))
    out_tensor = unsorted_segment_sum(data, segment_ids, num_segments)

    return out_tensor


def chunkify(a: torch.Tensor, dim: int, maxnumel: int) -> \
        Generator[Tuple[torch.Tensor, int, int], None, None]:
    """Splits the tensor `a` into several chunks of size `maxnumel` along the
    dimension given by `dim`.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.pytorch_utils import chunkify
    >>> a = torch.arange(10)
    >>> for chunk, istart, iend in chunkify(a, 0, 3):
    ...     print(chunk, istart, iend)
    tensor([0, 1, 2]) 0 3
    tensor([3, 4, 5]) 3 6
    tensor([6, 7, 8]) 6 9
    tensor([9]) 9 12

    Parameters
    ----------
    a: torch.Tensor
        The big tensor to be splitted into chunks.
    dim: int
        The dimension where the tensor would be splitted.
    maxnumel: int
        Maximum number of elements in a chunk.

    Returns
    -------
    chunks: Generator[Tuple[torch.Tensor, int, int], None, None]
        A generator that yields a tuple of three elements: the chunk tensor, the
        starting index of the chunk and the ending index of the chunk.

    """
    dim = a.ndim + dim if dim < 0 else dim

    numel = a.numel()
    dimnumel = a.shape[dim]
    nondimnumel = numel // dimnumel
    if maxnumel < nondimnumel:
        msg = "Cannot split the tensor of shape %s along dimension %s with maxnumel %d" % \
              (a.shape, dim, maxnumel)
        raise RuntimeError(msg)

    csize = min(maxnumel // nondimnumel, dimnumel)
    ioffset = 0
    lslice = (slice(None, None, None),) * dim
    rslice = (slice(None, None, None),) * (a.ndim - dim - 1)
    while ioffset < dimnumel:
        iend = ioffset + csize
        chunks = a[(lslice + (slice(ioffset, iend, None),) +
                    rslice)], ioffset, iend
        yield chunks
        ioffset = iend


def get_memory(a: torch.Tensor) -> int:
    """Returns the size of the tensor in bytes.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.pytorch_utils import get_memory
    >>> a = torch.randn(100, 100, dtype=torch.float64)
    >>> get_memory(a)
    80000

    Parameters
    ----------
    a: torch.Tensor
        The tensor to be measured.

    Returns
    -------
    size: int
        The size of the tensor in bytes.

    """
    size = a.numel() * a.element_size()
    return size


def gaussian_integral(
        n: int, alpha: Union[float,
                             torch.Tensor]) -> Union[float, torch.Tensor]:
    """Performs the gaussian integration.

    Examples
    --------
    >>> gaussian_int(5, 1.0)
    1.0

    Parameters
    ----------
    n: int
        The order of the integral
    alpha: Union[float, torch.Tensor]
        The parameter of the gaussian

    Returns
    -------
    Union[float, torch.Tensor]
        The value of the integral

    """
    n1 = (n + 1) * 0.5
    return scipy.special.gamma(n1) / (2 * alpha**n1)



# flake8: noqa
try:
    from deepchem.utils.differentiation_utils.editable_module import EditableModule

    from deepchem.utils.differentiation_utils.bcast import normalize_bcast_dims
    from deepchem.utils.differentiation_utils.bcast import get_bcasted_dims
    from deepchem.utils.differentiation_utils.bcast import match_dim

    from deepchem.utils.differentiation_utils.misc import set_default_option
    from deepchem.utils.differentiation_utils.misc import get_and_pop_keys
    from deepchem.utils.differentiation_utils.misc import get_method
    from deepchem.utils.differentiation_utils.misc import dummy_context_manager
    from deepchem.utils.differentiation_utils.misc import assert_runtime

    from deepchem.utils.differentiation_utils.linop import LinearOperator
    from deepchem.utils.differentiation_utils.linop import AddLinearOperator
    from deepchem.utils.differentiation_utils.linop import MulLinearOperator
    from deepchem.utils.differentiation_utils.linop import AdjointLinearOperator
    from deepchem.utils.differentiation_utils.linop import MatmulLinearOperator
    from deepchem.utils.differentiation_utils.linop import MatrixLinearOperator
except:
    pass



import itertools
from typing import Callable, List, Union, Literal, Optional

import dgl
import dgl.function as fn
import torch
from torch import nn
from torch.nn import functional as F

from deepchem.models.losses import NTXentMultiplePositives
from deepchem.models.torch_models import ModularTorchModel
from deepchem.models.torch_models.layers import MultilayerPerceptron
from deepchem.models.torch_models.pna_gnn import PNA, AtomEncoder
from deepchem.utils.graph_utils import fourier_encode_dist


class Net3DLayer(nn.Module):
    """
    Net3DLayer is a single layer of a 3D graph neural network based on the 3D Infomax architecture [1].

    This class expects a DGL graph with node features stored under the name 'feat' and edge features stored under the name 'd' (representing 3D distances). The edge features are updated by the message network and the node features are updated by the update network.

    Parameters
    ----------
    edge_dim : int
        The dimension of the edge features.
    hidden_dim : int
        The dimension of the hidden layers.
    reduce_func : str
        The reduce function to use for aggregating messages. Can be either 'sum' or 'mean'.
    batch_norm : bool, optional (default=False)
        Whether to use batch normalization.
    batch_norm_momentum : float, optional (default=0.1)
        The momentum for the batch normalization layers.
    dropout : float, optional (default=0.0)
        The dropout rate for the layers.
    mid_activation : str, optional (default='SiLU')
        The activation function to use in the network.
    message_net_layers : int, optional (default=2)
        The number of message network layers.
    update_net_layers : int, optional (default=2)
        The number of update network layers.

    References
    ----------
    .. [1] Stärk, H. et al. 3D Infomax improves GNNs for Molecular Property Prediction. Preprint at https://doi.org/10.48550/arXiv.2110.04126 (2022).

    Examples
    --------
    >>> net3d_layer = Net3DLayer(edge_dim=3, hidden_dim=3)
    >>> graph = dgl.graph(([0, 1], [1, 2]))
    >>> graph.ndata['feat'] = torch.tensor([[1., 2., 3.], [4., 5., 6.], [7., 8., 9.]])
    >>> graph.edata['d'] = torch.tensor([[0.5, 0.6, 0.7], [0.8, 0.9, 1.0]])
    >>> output = net3d_layer(graph)
    """

    def __init__(self,
                 edge_dim: int,
                 hidden_dim: int,
                 reduce_func: str = 'sum',
                 batch_norm: bool = False,
                 batch_norm_momentum: float = 0.1,
                 dropout: float = 0.0,
                 message_net_layers: int = 2,
                 update_net_layers: int = 2):
        super(Net3DLayer, self).__init__()

        self.message_network = nn.Sequential(
            MultilayerPerceptron(d_input=hidden_dim * 2 + edge_dim,
                                 d_output=hidden_dim,
                                 d_hidden=(hidden_dim,) *
                                 (message_net_layers - 1),
                                 batch_norm=batch_norm,
                                 batch_norm_momentum=batch_norm_momentum,
                                 dropout=dropout), torch.nn.SiLU())
        if reduce_func == 'sum':
            self.reduce_func = fn.sum
        elif reduce_func == 'mean':
            self.reduce_func = fn.mean
        else:
            raise ValueError('reduce function not supported: ', reduce_func)

        self.update_network = MultilayerPerceptron(
            d_input=hidden_dim,
            d_hidden=(hidden_dim,) * (update_net_layers - 1),
            d_output=hidden_dim,
            batch_norm=True,
            batch_norm_momentum=batch_norm_momentum)

        self.soft_edge_network = nn.Linear(hidden_dim, 1)

    def forward(self, input_graph: dgl.DGLGraph):
        """Perform a forward pass on the given graph.

        Parameters
        ----------
        input_graph : dgl.DGLGraph
            The graph to perform the forward pass on.

        Returns
        -------
        dgl.DGLGraph
            The updated graph after the forward pass.
        """
        # copy the input graph to avoid in-place operations
        graph = input_graph.local_var()
        graph.ndata['feat'] = input_graph.ndata['feat'].clone()
        graph.edata['d'] = input_graph.edata['d'].clone()

        graph.update_all(message_func=self.message_function,
                         reduce_func=self.reduce_func(msg='m', out='m_sum'),
                         apply_node_func=self.update_function)
        return graph

    def message_function(self, edges):
        """Computes the message and edge weight for a given set of edges.

        Parameters
        ----------
        edges : dgl.EdgeBatch
            A dgl.EdgeBatch object containing the edges information (data, batch size, etc.).

        Returns
        -------
        dict
            A dictionary containing the message multiplied by the edge weight.
        """
        message_input = torch.cat(
            [edges.src['feat'], edges.dst['feat'], edges.data['d']], dim=-1)
        message = self.message_network(message_input)
        edges.data['d'] += message
        edge_weight = torch.sigmoid(self.soft_edge_network(message))
        return {'m': message * edge_weight}

    def update_function(self, nodes):
        """
        Update function for updating node features based on the aggregated messages.

        This function is used in the forward method to perform a forward pass on the graph.

        Parameters
        ----------
        nodes : dgl.NodeBatch
            A node batch object containing the nodes information (data, batch size, etc.).

        Returns
        -------
        dict
            A dictionary containing the updated features.
        """
        h = nodes.data['feat']
        input = torch.cat([nodes.data['m_sum'] + nodes.data['feat']], dim=-1)
        h_new = self.update_network(input)
        output = h_new + h
        return {'feat': output}


class Net3D(nn.Module):
    """
    Net3D is a 3D graph neural network that expects a DGL graph input with 3D coordinates stored under the name 'd' and node features stored under the name 'feat'. It is based on the 3D Infomax architecture [1].

    Parameters
    ----------
    hidden_dim : int
        The dimension of the hidden layers.
    target_dim : int
        The dimension of the output layer.
    readout_aggregators : List[str]
        A list of aggregator functions for the readout layer. Options are 'sum', 'max', 'min', 'mean'.
    batch_norm : bool, optional (default=False)
        Whether to use batch normalization.
    node_wise_output_layers : int, optional (default=2)
        The number of output layers for each node.
    readout_batchnorm : bool, optional (default=True)
        Whether to use batch normalization in the readout layer.
    batch_norm_momentum : float, optional (default=0.1)
        The momentum for the batch normalization layers.
    reduce_func : str, optional (default='sum')
        The reduce function to use for aggregating messages.
    dropout : float, optional (default=0.0)
        The dropout rate for the layers.
    propagation_depth : int, optional (default=4)
        The number of propagation layers in the network.
    readout_layers : int, optional (default=2)
        The number of readout layers in the network.
    readout_hidden_dim : int, optional (default=None)
        The dimension of the hidden layers in the readout network.
    fourier_encodings : int, optional (default=0)
        The number of Fourier encodings to use.
    activation : str, optional (default='SiLU')
        The activation function to use in the network.
    update_net_layers : int, optional (default=2)
        The number of update network layers.
    message_net_layers : int, optional (default=2)
        The number of message network layers.
    use_node_features : bool, optional (default=False)
        Whether to use node features as input.

    Examples
    --------
    >>> from deepchem.feat.molecule_featurizers.conformer_featurizer import RDKitConformerFeaturizer
    >>> from deepchem.models.torch_models.gnn3d import Net3D
    >>> smiles = ["C[C@H](F)Cl", "C[C@@H](F)Cl"]
    >>> featurizer = RDKitConformerFeaturizer()
    >>> data = featurizer.featurize(smiles)
    >>> dgldata = [[graph.to_dgl_graph() for graph in conf] for conf in data]
    >>> net3d = Net3D(hidden_dim=3, target_dim=2, readout_aggregators=['sum', 'mean'])
    >>> output = [[net3d(graph) for graph in conf] for conf in dgldata]

    References
    ----------
    .. [1] Stärk, H. et al. 3D Infomax improves GNNs for Molecular Property Prediction. Preprint at https://doi.org/10.48550/arXiv.2110.04126 (2022).
    """

    def __init__(self,
                 hidden_dim,
                 target_dim,
                 readout_aggregators: List[str],
                 node_wise_output_layers=2,
                 batch_norm=True,
                 batch_norm_momentum=0.1,
                 reduce_func='sum',
                 dropout=0.0,
                 propagation_depth: int = 4,
                 readout_layers: int = 2,
                 readout_hidden_dim=None,
                 fourier_encodings=4,
                 update_net_layers=2,
                 message_net_layers=2,
                 use_node_features=False):
        super(Net3D, self).__init__()
        self.fourier_encodings = fourier_encodings
        edge_in_dim = 1 if fourier_encodings == 0 else 2 * fourier_encodings + 1

        self.edge_input = nn.Sequential(
            MultilayerPerceptron(d_input=edge_in_dim,
                                 d_output=hidden_dim,
                                 d_hidden=(hidden_dim,),
                                 batch_norm=batch_norm,
                                 batch_norm_momentum=batch_norm_momentum),
            torch.nn.SiLU())

        self.use_node_features = use_node_features
        if self.use_node_features:
            self.atom_encoder = AtomEncoder(hidden_dim)
        else:
            self.node_embedding = nn.Parameter(torch.empty((hidden_dim,)))
            nn.init.normal_(self.node_embedding)

        self.mp_layers = nn.ModuleList()
        for _ in range(propagation_depth):
            self.mp_layers.append(
                Net3DLayer(edge_dim=hidden_dim,
                           hidden_dim=hidden_dim,
                           batch_norm=batch_norm,
                           batch_norm_momentum=batch_norm_momentum,
                           dropout=dropout,
                           reduce_func=reduce_func,
                           message_net_layers=message_net_layers,
                           update_net_layers=update_net_layers))

        self.node_wise_output_layers = node_wise_output_layers
        if self.node_wise_output_layers > 0:
            self.node_wise_output_network = MultilayerPerceptron(
                d_input=hidden_dim,
                d_output=hidden_dim,
                d_hidden=(hidden_dim,),
                batch_norm=batch_norm,
                batch_norm_momentum=batch_norm_momentum)

        if readout_hidden_dim is None:
            readout_hidden_dim = hidden_dim
        self.readout_aggregators = readout_aggregators

        self.output = MultilayerPerceptron(
            d_input=hidden_dim * len(self.readout_aggregators),
            d_output=target_dim,
            d_hidden=(readout_hidden_dim,) *
            (readout_layers -
             1),  # -1 because the input layer is not considered a hidden layer
            batch_norm=False)

    def forward(self, graph: dgl.DGLGraph):
        """
        Forward pass of the Net3D model.

        Parameters
        ----------
        graph : dgl.DGLGraph
            The input graph with node features stored under the key 'x' and edge distances stored under the key 'd'.

        Returns
        -------
        torch.Tensor
            The graph representation tensor of shape (1, target_dim).
        """
        if self.use_node_features:
            graph.ndata['feat'] = self.atom_encoder(graph.ndata['x'])
        else:
            graph.ndata['feat'] = self.node_embedding[None, :].expand(
                graph.number_of_nodes(), -1)

        if self.fourier_encodings > 0:
            graph.edata['d'] = fourier_encode_dist(
                graph.edata['d'], num_encodings=self.fourier_encodings)
        graph.apply_edges(self.input_edge_func)

        for mp_layer in self.mp_layers:
            graph = mp_layer(graph)

        if self.node_wise_output_layers > 0:
            graph.apply_nodes(self.output_node_func)

        readouts_to_cat = [
            dgl.readout_nodes(graph, 'feat', op=aggr)
            for aggr in self.readout_aggregators
        ]
        readout = torch.cat(readouts_to_cat, dim=-1)
        return self.output(readout.squeeze())

    def output_node_func(self, nodes):
        """
        Apply the node-wise output network to the node features.

        Parameters
        ----------
        nodes : dgl.NodeBatch
            A batch of nodes with features stored under the key 'feat'.

        Returns
        -------
        dict
            A dictionary with the updated node features under the key 'feat'.
        """
        return {'feat': self.node_wise_output_network(nodes.data['feat'])}

    def input_edge_func(self, edges):
        """
        Apply the edge input network to the edge features.

        Parameters
        ----------
        edges : dgl.EdgeBatch
            A batch of edges with distances stored under the key 'd'.

        Returns
        -------
        dict
            A dictionary with the updated edge features under the key 'd'.
        """
        return {'d': F.silu(self.edge_input(edges.data['d']))}


class InfoMax3DModular(ModularTorchModel):
    """
    InfoMax3DModular is a modular torch model that uses a 2D PNA model and a 3D Net3D model to maximize the mutual information between their representations. The 2D model can then be used for downstream tasks without the need for 3D coordinates. This is based off the work in [1].

    This class expects data in featurized by the RDKitConformerFeaturizer. This featurizer produces features of the type Array[Array[List[GraphData]]].
    The outermost array is the dataset array, the second array is the molecule, the list contains the conformers for that molecule and the GraphData object is the featurized graph for that conformer with node_pos_features holding the 3D coordinates.
    If you are not using RDKitConformerFeaturizer, your input data features should look like this: Dataset[Molecule[Conformers[GraphData]]].

    For pretraining, the original paper used a learning rate of 8e-5 with a batch size of 500.
    For finetuning on quantum mechanical datasets, a learning rate of 7e-5 with a batch size of 128
    was used. For finetuning on non-quantum mechanical datasets, a learning rate of 1e-3 with a
    batch size of 32 was used in the original implementation.

    Parameters
    ----------
    task : Literal['pretrain', 'regression', 'classification']
        The task of the model
    hidden_dim : int, optional, default = 64
        The dimension of the hidden layers.
    target_dim : int, optional, default = 10
        The dimension of the output layer.
    aggregators : List[str]
        A list of aggregator functions for the PNA model. Options are 'mean', 'sum', 'min', 'max', 'std', 'var', 'moment3', 'moment4', 'moment5'.
    readout_aggregators : List[str]
        A list of aggregator functions for the readout layer. Options are 'sum', 'max', 'min', 'mean'.
    scalers : List[str]
        A list of scaler functions for the PNA model. Options are 'identity', 'amplification', 'attenuation'.
    residual : bool, optional (default=True)
        Whether to use residual connections in the PNA model.
    node_wise_output_layers : int, optional (default=2)
        The number of output layers for each node in the Net3D model.
    pairwise_distances : bool, optional (default=False)
        Whether to use pairwise distances in the PNA model.
    activation : Union[Callable, str], optional (default="relu")
        The activation function to use in the PNA model.
    reduce_func : str, optional (default='sum')
        The reduce function to use for aggregating messages in the Net3D model.
    batch_norm : bool, optional (default=True)
        Whether to use batch normalization in the PNA model.
    batch_norm_momentum : float, optional (default=0.1)
        The momentum for the batch normalization layers.
    propagation_depth : int, optional (default=5)
        The number of propagation layers in the PNA and Net3D models.
    dropout : float, optional (default=0.0)
        The dropout rate for the layers in the PNA and Net3D models.
    readout_layers : int, optional (default=2)
        The number of readout layers in the PNA and Net3D models.
    readout_hidden_dim : int, optional (default=None)
        The dimension of the hidden layers in the readout network.
    fourier_encodings : int, optional (default=4)
        The number of Fourier encodings to use in the Net3D model.
    update_net_layers : int, optional (default=2)
        The number of update network layers in the Net3D model.
    message_net_layers : int, optional (default=2)
        The number of message network layers in the Net3D model.
    use_node_features : bool, optional (default=False)
        Whether to use node features as input in the Net3D model.
    posttrans_layers : int, optional (default=1)
        The number of post-transformation layers in the PNA model.
    pretrans_layers : int, optional (default=1)
        The number of pre-transformation layers in the PNA model.
    kwargs : dict
        Additional keyword arguments.

    References
    ----------
    .. [1] Stärk, H. et al. 3D Infomax improves GNNs for Molecular Property Prediction. Preprint at https://doi.org/10.48550/arXiv.2110.04126 (2022).

    Examples
    --------
    >>> from deepchem.feat.graph_data import BatchGraphData
    >>> from deepchem.feat.molecule_featurizers.conformer_featurizer import RDKitConformerFeaturizer
    >>> from deepchem.models.torch_models.gnn3d import InfoMax3DModular
    >>> import numpy as np
    >>> import deepchem as dc
    >>> from deepchem.data.datasets import NumpyDataset
    >>> smiles = ["C[C@H](F)Cl", "C[C@@H](F)Cl"]
    >>> featurizer = RDKitConformerFeaturizer()
    >>> data = featurizer.featurize(smiles)
    >>> dataset = NumpyDataset(X=data)
    >>> model = InfoMax3DModular(hidden_dim=64,
    ...                          target_dim=10,
    ...                          aggregators=['max'],
    ...                          readout_aggregators=['mean'],
    ...                          scalers=['identity'])
    >>> loss = model.fit(dataset, nb_epoch=1)
    """

    def __init__(self,
                 task: Literal['pretraining', 'regression', 'classification'],
                 hidden_dim: int = 64,
                 target_dim: int = 10,
                 aggregators: List[str] = ['mean'],
                 readout_aggregators: List[str] = ['mean'],
                 scalers: List[str] = ['identity'],
                 residual: bool = True,
                 node_wise_output_layers: int = 2,
                 pairwise_distances: bool = False,
                 activation: Union[Callable, str] = "relu",
                 reduce_func: str = 'sum',
                 batch_norm: bool = True,
                 batch_norm_momentum: float = 0.1,
                 propagation_depth: int = 5,
                 dropout: float = 0.0,
                 readout_layers: int = 2,
                 readout_hidden_dim: int = 1,
                 fourier_encodings: int = 4,
                 update_net_layers: int = 2,
                 message_net_layers: int = 2,
                 use_node_features: bool = False,
                 posttrans_layers: int = 1,
                 pretrans_layers: int = 1,
                 n_tasks: int = 1,
                 n_classes: Optional[bool] = None,
                 **kwargs):
        self.task = task
        self.hidden_dim = hidden_dim
        self.target_dim = target_dim
        self.aggregators = aggregators
        self.readout_aggregators = readout_aggregators
        self.scalers = scalers
        self.residual = residual
        self.node_wise_output_layers = node_wise_output_layers
        self.pairwise_distances = pairwise_distances
        self.activation = activation
        self.reduce_func = reduce_func
        self.batch_norm = batch_norm
        self.batch_norm_momentum = batch_norm_momentum
        self.propagation_depth = propagation_depth
        self.dropout = dropout
        self.readout_layers = readout_layers
        self.readout_hidden_dim = readout_hidden_dim
        self.fourier_encodings = fourier_encodings
        self.update_net_layers = update_net_layers
        self.message_net_layers = message_net_layers
        self.use_node_features = use_node_features
        self.posttrans_layers = posttrans_layers
        self.pretrans_layers = pretrans_layers
        self.kwargs = kwargs
        self.n_tasks = n_tasks
        self.n_classes = n_classes
        self.criterion = NTXentMultiplePositives()._create_pytorch_loss()
        self.components = self.build_components()
        self.model = self.build_model()
        if self.task == 'regression':
            output_types = ['prediction']
        elif self.task == 'classification':
            output_types = ['prediction', 'logits']
        else:
            output_types = None
        super().__init__(self.model,
                         self.components,
                         output_types=output_types,
                         **kwargs)
        for module_name, module in self.components.items():
            self.components[module_name] = module.to(self.device)
        self.model = self.model.to(self.device)

    def build_components(self):
        """
        Build the components of the InfoMax3DModular model.

        Returns
        -------
        dict
            A dictionary containing the '2d' PNA model and the '3d' Net3D model.
        """
        model2d = PNA(hidden_dim=self.hidden_dim,
                      target_dim=self.target_dim,
                      aggregators=self.aggregators,
                      scalers=self.scalers,
                      readout_aggregators=self.readout_aggregators,
                      readout_hidden_dim=self.readout_hidden_dim,
                      readout_layers=self.readout_layers,
                      residual=self.residual,
                      pairwise_distances=self.pairwise_distances,
                      activation=self.activation,
                      batch_norm=self.batch_norm,
                      batch_norm_momentum=self.batch_norm_momentum,
                      propagation_depth=self.propagation_depth,
                      dropout=self.dropout,
                      posttrans_layers=self.posttrans_layers,
                      pretrans_layers=self.pretrans_layers,
                      task=self.task,
                      n_tasks=self.n_tasks,
                      n_classes=self.n_classes,
                      **self.kwargs)
        if self.task == 'pretraining':
            return {
                'model2d':
                    model2d,
                'model3d':
                    Net3D(hidden_dim=self.hidden_dim,
                          target_dim=self.target_dim,
                          readout_aggregators=self.readout_aggregators,
                          node_wise_output_layers=self.node_wise_output_layers,
                          batch_norm=True,
                          batch_norm_momentum=self.batch_norm_momentum,
                          reduce_func=self.reduce_func,
                          dropout=self.dropout,
                          propagation_depth=self.propagation_depth,
                          readout_layers=self.readout_layers,
                          readout_hidden_dim=self.readout_hidden_dim,
                          fourier_encodings=self.fourier_encodings,
                          update_net_layers=self.update_net_layers,
                          message_net_layers=self.message_net_layers,
                          use_node_features=self.use_node_features)
            }
        elif self.task in ['regression', 'classification']:
            return {'model2d': model2d}

    def build_model(self):
        """
        Build the InfoMax3DModular model. This is the 2D network which is meant to be used for inference.

        Returns
        -------
        PNA
            The 2D PNA model component.
        """
        # FIXME For pretraining task, both model2d and model3d but the super class
        # can't handle two models for contrastive learning, hence we pass only model2d
        return self.components['model2d']

    def loss_func(self, inputs, labels, weights):
        """
        Compute the loss function for the InfoMax3DModular model.

        Parameters
        ----------
        inputs : dgl.DGLGraph
            The input graph with node features stored under the key 'x' and edge distances stored under the key 'd'.
        labels : torch.Tensor
            The ground truth labels.
        weights : torch.Tensor
            The weights for each sample.

        Returns
        -------
        torch.Tensor
            The computed loss value.
        """
        if self.task == 'pretraining':
            encodings2d = self.components['model2d'](inputs)
            encodings3d = self.components['model3d'](inputs)
            loss = self.criterion(encodings2d, encodings3d)
        elif self.task == 'regression':
            preds = self.model(inputs)
            loss = F.mse_loss(preds, labels)
        elif self.task == 'classification':
            proba, logits = self.model(inputs)
            # torch's one-hot encoding works with integer data types.
            # We convert labels to integer, one-hot encode and convert it back to float
            # for making it suitable to loss function
            labels = F.one_hot(labels.squeeze().type(torch.int64)).type(
                torch.float32)
            loss = F.binary_cross_entropy_with_logits(logits, labels)
        return loss

    def _prepare_batch(self, batch):
        """
        Prepare a batch of data for the InfoMax3DModular model.

        Parameters
        ----------
        batch : tuple
            A tuple containing the inputs, labels, and weights.

        Returns
        -------
        tuple
            A tuple containing the prepared batch graph, labels, and weights.
        """
        inputs, labels, weights = batch
        inputs = inputs[0]
        if labels is not None:
            labels = torch.from_numpy(labels[0]).float().to(self.device)

        # graphs = [[
        #     graph_data.to_dgl_graph().to(self.device) for graph_data in row
        # ] for row in inputs]

        # convert the GraphData objects to DGL graphs
        graphs = dgl.batch([graph_data.to_dgl_graph() for graph_data in inputs
                           ]).to(self.device)
        return graphs, labels, weights

    def _ensure_built(self) -> None:
        """The first time this is called, create internal data structures."""
        if self._built:
            return
        self._built = True
        self._global_step = 0
        if self.task == 'pretraining':
            normal_params = [
                v for k, v in itertools.chain(
                    self.components['model2d'].named_parameters(),
                    self.components['model3d'].named_parameters())
                if 'batch_norm' not in k
            ]
            batch_norm_params = [
                v for k, v in itertools.chain(
                    self.components['model2d'].named_parameters(),
                    self.components['model3d'].named_parameters())
                if 'batch_norm' in k
            ]

        elif self.task in ['regression', 'classification']:
            normal_params = [
                v for k, v in self.model.named_parameters()
                if 'batch_norm' not in k
            ]
            batch_norm_params = [
                v for k, v in self.model.named_parameters() if 'batch_norm' in k
            ]

        params = [{
            'params': batch_norm_params,
            'weight_decay': 0
        }, {
            'params': normal_params
        }]
        self._pytorch_optimizer = torch.optim.Adam(
            params,  # type: ignore
            lr=self.learning_rate)  # type: ignore
        # TODO Ideally, we should use a lr schedule but we need to update lr_scheduler.step() method
        # in ModularTorchModel.fit_generator to accept a metric.
        self._lr_schedule = None
        # self._lr_schedule = torch.optim.lr_scheduler.ReduceLROnPlateau(self._pytorch_optimizer,
        #                                                            mode='min')



"""
BBBC Dataset loader.

This file contains image loaders for the BBBC dataset collection (https://data.broadinstitute.org/bbbc/image_sets.html).
"""
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union
import numpy as np
import pandas as pd

BBBC1_IMAGE_URL = 'https://data.broadinstitute.org/bbbc/BBBC001/BBBC001_v1_images_tif.zip'
BBBC1_LABEL_URL = 'https://data.broadinstitute.org/bbbc/BBBC001/BBBC001_v1_counts.txt'
BBBC1_TASKS = ["cell-count"]

BBBC2_IMAGE_URL = 'https://data.broadinstitute.org/bbbc/BBBC002/BBBC002_v1_images.zip'
BBBC2_LABEL_URL = 'https://data.broadinstitute.org/bbbc/BBBC002/BBBC002_v1_counts.txt'
BBBC2_TASKS = ["cell-count"]


class _BBBC001Loader(_MolnetLoader):

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, "BBBC001_v1_images_tif.zip")
        labels_file = os.path.join(self.data_dir, "BBBC001_v1_counts.txt")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=BBBC1_IMAGE_URL,
                                             dest_dir=self.data_dir)
        if not os.path.exists(labels_file):
            dc.utils.data_utils.download_url(url=BBBC1_LABEL_URL,
                                             dest_dir=self.data_dir)
        labels_table = pd.read_csv(labels_file, delimiter="\t")
        labels = np.mean(
            [labels_table["manual count #1"], labels_table["manual count #2"]],
            axis=0,
            dtype=int)

        loader = dc.data.ImageLoader()
        return loader.create_dataset(inputs=(dataset_file, labels),
                                     in_memory=False)


def load_bbbc001(
    splitter: Union[dc.splits.Splitter, str, None] = 'index',
    transformers: List[Union[TransformerGenerator, str]] = [],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load BBBC001 dataset

    This dataset contains 6 images of human HT29 colon cancer cells. The task is
    to learn to predict the cell counts in these images. This dataset is too small
    to serve to train algorithms, but might serve as a good test dataset.
    https://data.broadinstitute.org/bbbc/BBBC001/

    Parameters
    ----------
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in
    """
    featurizer = dc.feat.UserDefinedFeaturizer([])  # Not actually used
    loader = _BBBC001Loader(featurizer, splitter, transformers, BBBC1_TASKS,
                            data_dir, save_dir, **kwargs)
    return loader.load_dataset('bbbc001', reload)


class _BBBC002Loader(_MolnetLoader):

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, "BBBC002_v1_images.zip")
        labels_file = os.path.join(self.data_dir, "BBBC002_v1_counts.txt.txt")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=BBBC2_IMAGE_URL,
                                             dest_dir=self.data_dir)
        if not os.path.exists(labels_file):
            dc.utils.data_utils.download_url(url=BBBC2_LABEL_URL,
                                             dest_dir=self.data_dir)

        labels_table = pd.read_csv(labels_file, delimiter="\t")
        labels = np.mean([
            labels_table["human counter 1 (Robert Lindquist)"],
            labels_table["human counter #2 (Joohan Chang)"]
        ],
                         axis=0,
                         dtype=int)

        loader = dc.data.ImageLoader()
        return loader.create_dataset(inputs=(dataset_file, labels),
                                     in_memory=False)


def load_bbbc002(
    splitter: Union[dc.splits.Splitter, str, None] = 'index',
    transformers: List[Union[TransformerGenerator, str]] = [],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load BBBC002 dataset

    This dataset contains data corresponding to 5 samples of Drosophilia Kc167
    cells. There are 10 fields of view for each sample, each an image of size
    512x512. Ground truth labels contain cell counts for this dataset. Full
    details about this dataset are present at
    https://data.broadinstitute.org/bbbc/BBBC002/.

    Parameters
    ----------
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in
    """
    featurizer = dc.feat.UserDefinedFeaturizer([])  # Not actually used
    loader = _BBBC002Loader(featurizer, splitter, transformers, BBBC2_TASKS,
                            data_dir, save_dir, **kwargs)
    return loader.load_dataset('bbbc002', reload)
